import json
import smtplib
import traceback
from collections import defaultdict
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List

import pendulum
from sqlalchemy.orm.exc import NoResultFound

from riberry import model, config
from celery.utils.log import logger


def email_notification(host, body, mime_type, subject, sender, recipients: List):
    try:
        msg = MIMEMultipart('alternative')
        msg['Subject'] = subject
        msg['From'] = sender
        msg['To'] = ', '.join(recipients)
        msg.attach(MIMEText(body, mime_type))
        s = smtplib.SMTP(host)
        s.sendmail(sender, recipients, msg.as_string())
        s.quit()
    except:
        logger.warn(f'An error occurred while sending email notification: {traceback.format_exc()}')


def handle_artifacts(events: List[model.misc.Event]):
    job_executions = {}
    streams = {}

    to_delete = []
    for event in events:
        try:
            event_data = json.loads(event.data)
            stream_name = event_data['stream']

            artifact = model.job.JobExecutionArtifact(
                name=event_data['name'] or 'Untitled',
                type=model.job.ArtifactType[event_data['type']],
                category=event_data['category'] or 'Default',
                filename=event_data['filename'] or 'Untitled',
                size=len(event.binary) if event.binary else 0,
                created=pendulum.from_timestamp(event.time),
                binary=model.job.JobExecutionArtifactBinary(binary=event.binary),
                data=[
                    model.job.JobExecutionArtifactData(
                        title=str(title),
                        description=str(description)
                    ) for title, description in event_data['data'].items() if title and description
                ]
            )

            if event.root_id not in job_executions:
                try:
                    job_executions[event.root_id] = model.job.JobExecution.query().filter_by(task_id=event.root_id).one()
                except NoResultFound:
                    to_delete.append(event)
                    continue

            job_execution = job_executions[event.root_id]
            artifact.job_execution = job_executions[event.root_id]

            if stream_name:
                if (stream_name, event.root_id) not in streams:
                    try:
                        streams[(stream_name, event.root_id)] = model.job.JobExecutionStream.query().filter_by(
                            name=stream_name, job_execution=job_execution).one()
                    except NoResultFound:
                        to_delete.append(event)
                        continue
                stream = streams[(stream_name, event.root_id)]
                artifact.stream = stream

            model.conn.add(artifact)
        except:
            logger.warn(f'An error occurred processing artifact event {event}: {traceback.format_exc()}')
        else:
            to_delete.append(event)

    return to_delete


def handle_steps(events: List[model.misc.Event]):
    to_delete = []
    steps = {}
    streams = {}
    job_executions = {}

    for event in events:
        try:
            event_data = json.loads(event.data)
            event_time = pendulum.from_timestamp(event.time)
            stream_name = event_data['stream']

            if event.root_id not in job_executions:
                try:
                    job_executions[event.root_id] = model.job.JobExecution.query().filter_by(
                        task_id=event.root_id).one()
                except NoResultFound:
                    to_delete.append(event)
                    continue
            job_execution = job_executions[event.root_id]

            if (stream_name, event.root_id) not in streams:
                try:
                    streams[(stream_name, event.root_id)] = model.job.JobExecutionStream.query().filter_by(
                        name=stream_name, job_execution=job_execution).one()
                except NoResultFound:
                    to_delete.append(event)
                    continue
            stream = streams[(stream_name, event.root_id)]

            try:
                if (stream_name, event.task_id) not in steps:
                    steps[(stream_name, event.task_id)] = model.job.JobExecutionStreamStep.query().filter_by(
                        task_id=event.task_id, stream=stream).one()
                step = steps[(stream_name, event.task_id)]
            except NoResultFound:
                step = model.job.JobExecutionStreamStep(
                    name=event_data['step'],
                    created=pendulum.from_timestamp(event.time),
                    updated=pendulum.from_timestamp(event.time),
                    task_id=event.task_id,
                    stream=stream,
                    status=event_data['state']
                )
                model.conn.add(step)

            step_updated = pendulum.instance(step.updated, tz='utc')
            if event_time >= step_updated:
                status = event_data['state']
                step.status = status
                step.updated = event_time
                if status == 'ACTIVE':
                    step.started = event_time
                elif status in ('SUCCESS', 'FAILURE'):
                    step.completed = event_time

        except:
            logger.warn(f'An error occurred processing step event {event}: {traceback.format_exc()}')
        else:
            to_delete.append(event)

    return to_delete


def handle_streams(events: List[model.misc.Event]):
    to_delete = []
    streams = {}
    job_executions = {}

    for event in events:
        try:
            event_data = json.loads(event.data)
            event_time = pendulum.from_timestamp(event.time, tz='utc')
            stream_name = event_data['stream']

            if not stream_name:
                logger.warn('Empty stream name provided, skipping')
                to_delete.append(event)
                continue

            if event.root_id not in job_executions:
                try:
                    job_executions[event.root_id] = model.job.JobExecution.query().filter_by(
                        task_id=event.root_id).one()
                except NoResultFound:
                    to_delete.append(event)
                    continue
            job_execution = job_executions[event.root_id]
            try:
                if (stream_name, event.root_id) not in streams:
                    streams[(stream_name, event.root_id)] = model.job.JobExecutionStream.query().filter_by(
                        name=stream_name, job_execution=job_execution).one()
                stream = streams[(stream_name, event.root_id)]
            except NoResultFound:
                existing_stream = model.job.JobExecutionStream.query().filter_by(task_id=event.task_id).first()
                if existing_stream:
                    logger.warn(f'Skipping stream event {event}. Task ID {event.task_id!r} already exists against '
                                f'an existing stream (id={existing_stream.id}).\n'
                                f'Details:\n'
                                f'  root_id: {event.root_id!r}\n'
                                f'  name: {stream_name!r}\n'
                                f'  data: {event_data}\n')
                    to_delete.append(event)
                    continue

                stream = model.job.JobExecutionStream(
                    name=str(event_data['stream']),
                    task_id=event.task_id,
                    created=pendulum.from_timestamp(event.time, tz='utc'),
                    updated=pendulum.from_timestamp(event.time, tz='utc'),
                    status='QUEUED',
                    job_execution=job_execution
                )
                model.conn.add(stream)

            stream_updated = pendulum.instance(stream.updated, tz='utc')
            if event_time >= stream_updated:
                status = event_data['state']
                stream.status = status
                stream.updated = event_time
                if status == 'ACTIVE' and stream.started is None:
                    stream.started = event_time
                elif status in ('SUCCESS', 'FAILURE'):
                    stream.completed = event_time
        except:
            logger.warn(f'An error occurred processing stream event {event}: {traceback.format_exc()}')
        else:
            to_delete.append(event)

    return to_delete


def handle_notifications(events: List[model.misc.Event]):
    to_delete = []

    for event in events:
        event_data = json.loads(event.data)
        notification_type = event_data['type']
        notification_data = event_data['data']

        try:
            execution: model.job.JobExecution = model.job.JobExecution.query().filter_by(task_id=event.root_id).one()
            user = execution.creator
        except NoResultFound:
            to_delete.append(event)
            continue

        if notification_type == 'custom-email' and config.config.email.enabled:
            try:
                email_notification(
                    host=config.config.email.smtp_server,
                    body=notification_data['body'],
                    mime_type=notification_data.get('mime_type') or 'plain',
                    subject=notification_data['subject'],
                    sender=notification_data.get('from') or config.config.email.sender,
                    recipients=[user.details.email] + notification_data.get('to', []),
                )
            except:
                logger.warn(f'An error occurred processing notification type {notification_type}: '
                            f'{traceback.format_exc()}')

        elif notification_type == 'workflow_complete':
            status = str(notification_data['status']).lower()
            message = f'Completed execution #{execution.id} for job ' \
                      f'{execution.job.name} with status {str(status).lower()}'
            notification = model.misc.Notification(
                type=(
                    model.misc.NotificationType.success if str(status).lower() == 'success'
                    else model.misc.NotificationType.error
                ),
                message=message,
                user_notifications=[
                    model.misc.UserNotification(user=user)
                ],
                targets=[
                    model.misc.NotificationTarget(target='JobExecution', target_id=execution.id)
                ]
            )
            model.conn.add(notification)
            if config.config.email.enabled:
                email_notification(
                    host=config.config.email.smtp_server,
                    body=message,
                    mime_type='plain',
                    subject=f'Riberry / {status.title()} / {execution.job.name} / execution #{execution.id}',
                    sender=config.config.email.sender,
                    recipients=[user.details.email],
                )

        elif notification_type == 'workflow_started':
            message = f'Processing execution #{execution.id} for job {execution.job.name}'
            notification = model.misc.Notification(
                type=model.misc.NotificationType.info,
                message=message,
                user_notifications=[
                    model.misc.UserNotification(user=execution.creator)
                ],
                targets=[
                    model.misc.NotificationTarget(target='JobExecution', target_id=execution.id)
                ]
            )
            model.conn.add(notification)
            if config.config.email.enabled:
                email_notification(
                    host=config.config.email.smtp_server,
                    body=message,
                    mime_type='plain',
                    subject=f'Riberry / Started / {execution.job.name} / execution #{execution.id}',
                    sender=config.config.email.sender,
                    recipients=[user.details.email],
                )
        else:
            logger.warn(f'Received unknown notification type {notification_type}')

        to_delete.append(event)

    return to_delete


handlers = {
    'stream': handle_streams,
    'step': handle_steps,
    'artifact': handle_artifacts,
    'notify': handle_notifications,
}


def process(event_limit=None):
    with model.conn:
        event_mapping = defaultdict(list)
        query = model.misc.Event.query().order_by(model.misc.Event.time.asc(), model.misc.Event.id.asc())
        if event_limit:
            query = query.limit(event_limit)
        events = query.all()

        if not events:
            return

        for event in events:
            event_mapping[event.name].append(event)

        to_delete = []
        for handler_name, handler_func in handlers.items():
            handler_events = event_mapping[handler_name]
            if handler_events:
                try:
                    to_delete += handlers[handler_name](handler_events)
                except:
                    logger.warn(f'Failed to process {handler_name} events: {handler_events}')
                    raise

        for event in to_delete:
            logger.info(f'Removing processed event {event}')
            model.conn.delete(event)

        model.conn.commit()


if __name__ == '__main__':
    process()



import riberry


class RiberryApplicationConfig:

    def __init__(self, **kwargs):
        self.enable_steps = kwargs.get('enable_steps', True)


class RiberryApplication:
    __registered__ = {}

    def __init__(self, *, name, backend, config=None, addons=None):
        self.name = name
        self.config = config or RiberryApplicationConfig()
        self.__registered__[self.name] = self
        self.context: riberry.app.context.Context = riberry.app.context.Context()
        self.entry_points = {}
        self.backend: riberry.app.backends.RiberryApplicationBackend = backend
        self.backend.initialize()
        self.addons = {
            **self.backend.default_addons(),
            **(addons or {})
        }
        for addon in self.addons.values():
            if addon is not None:
                addon.register(riberry_app=self)

    @classmethod
    def by_name(cls, name):
        return cls.__registered__[name]

    def entry_point(self, form, stream='Overall', step='Entry'):
        def wrapper(func):
            self.entry_points[form] = EntryPoint(
                form=form,
                func=func,
                stream=stream,
                step=step,
            )

        return wrapper

    def task(self, func=None, **options):
        return self.backend.task(func=func, **options)

    def register_task(self, func, **options):
        return self.backend.register_task(func=func, **options)

    def start(self, execution_id, root_id, form) -> str:
        if form not in self.entry_points:
            raise ValueError(f'Application {self.name!r} does not have an entry point with '
                             f'name {form!r} registered.')

        entry_point: EntryPoint = self.entry_points[form]
        with self.context.flow.stream_scope(stream=entry_point.stream):
            return self.backend.start_execution(execution_id=execution_id, root_id=root_id, entry_point=entry_point)


class EntryPoint:

    def __init__(self, form, func, stream, step):
        self.form = form
        self.func = func
        self.stream = stream
        self.step = step



import os

import riberry
from .base import RiberryApplication
from .context import Context
from .util.misc import Proxy

__cache = dict(
    app_name={},
)
__cache_app_name = __cache['app_name']


def get_instance_name(raise_on_none=True) -> str:
    if 'RIBERRY_INSTANCE' not in os.environ and raise_on_none:
        raise EnvironmentError("Environment variable 'RIBERRY_INSTANCE' not set")
    return os.environ.get('RIBERRY_INSTANCE')


def get_application_name() -> str:
    instance_name = get_instance_name(raise_on_none=True)
    if instance_name not in __cache_app_name:
        __cache_app_name[instance_name] = get_instance_model().application.internal_name
    return __cache_app_name[instance_name]


def get_instance_model() -> riberry.model.application.ApplicationInstance:
    return riberry.model.application.ApplicationInstance.query().filter_by(
        internal_name=get_instance_name(raise_on_none=True),
    ).one()


def is_current_instance(instance_name: str) -> bool:
    return bool(instance_name and get_instance_name(raise_on_none=False) == instance_name)


current_riberry_app: RiberryApplication = Proxy(
    getter=lambda: RiberryApplication.by_name(name=get_application_name())
)
current_context: Context = Proxy(
    getter=lambda: current_riberry_app.context
)



from . import defaults
import importlib
import pkgutil
from collections import defaultdict


plugin_register = defaultdict(set)
plugin_register['authentication'].add(defaults.authentication.DefaultAuthenticationProvider)
plugin_register['policies'].add(defaults.policies.default_policies)


ext_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg in pkgutil.iter_modules()
    if name.startswith('riberry_')
}



import enum
from typing import List

from sqlalchemy import String, Column, Enum, ForeignKey, sql
from sqlalchemy.orm import relationship

from riberry import model
from riberry.model import base


class ResourceGroupAssociation(base.Base):
    __tablename__ = 'resource_group'
    __reprattrs__ = ['group_id', 'resource_id', 'resource_type']

    # columns
    id = base.id_builder.build()
    group_id = Column(ForeignKey('groups.id'), nullable=False)
    resource_id = Column(base.id_builder.type, nullable=False)
    resource_type = Column(Enum(model.misc.ResourceType), nullable=False)

    # associations
    group: 'Group' = relationship('Group', back_populates='resource_associations')

    @classmethod
    def make_relationship(cls, resource_id, resource_type):
        return relationship(
            'ResourceGroupAssociation',
            primaryjoin=lambda: sql.and_(
                resource_id == ResourceGroupAssociation.resource_id,
                ResourceGroupAssociation.resource_type == resource_type
            ),
            foreign_keys=lambda: ResourceGroupAssociation.resource_id,
            cascade='save-update, merge, delete, delete-orphan',
        )


class Group(base.Base):
    __tablename__ = 'groups'
    __reprattrs__ = ['name']

    # columns
    id = base.id_builder.build()
    name: str = Column(String(128), nullable=False, unique=True)
    _display_name: str = Column('display_name', String(128))
    description: str = Column(String(128))

    # associations
    resource_associations: List['ResourceGroupAssociation'] = relationship(
        'ResourceGroupAssociation', back_populates='group')
    user_associations: List['ResourceGroupAssociation'] = relationship(
        'ResourceGroupAssociation',
        primaryjoin=lambda: sql.and_(
            ResourceGroupAssociation.group_id == Group.id,
            ResourceGroupAssociation.resource_type == model.misc.ResourceType.user
        )
    )
    form_associations: List['ResourceGroupAssociation'] = relationship(
        'ResourceGroupAssociation',
        primaryjoin=lambda: sql.and_(
            ResourceGroupAssociation.group_id == Group.id,
            ResourceGroupAssociation.resource_type == model.misc.ResourceType.form
        )
    )

    @property
    def display_name(self):
        return self._display_name or self.name

    @property
    def users(self):
        return model.auth.User.query().filter(
            (ResourceGroupAssociation.group_id == self.id) &
            (ResourceGroupAssociation.resource_type == model.misc.ResourceType.user) &
            (model.auth.User.id == ResourceGroupAssociation.resource_id)
        ).all()

    @property
    def forms(self):
        return model.interface.Form.query().filter(
            (ResourceGroupAssociation.group_id == self.id) &
            (ResourceGroupAssociation.resource_type == model.misc.ResourceType.form) &
            (model.interface.Form.id == ResourceGroupAssociation.resource_id)
        ).all()







import json
import sys
import traceback
from typing import Union, Optional

import riberry
from riberry.exc import BaseError
from riberry.model.job import ArtifactType
from .. import current_context as context
from ..util.events import create_event


def create_artifact(filename: str, content: Union[bytes, str], name: str = None,
                    type: Union[str, ArtifactType] = ArtifactType.output, category='Default', data: dict = None,
                    stream: str = None, task_id: str = None, root_id: str = None):
    task_id = task_id or context.current.task_id
    root_id = root_id or context.current.root_id
    stream = stream or context.current.stream

    if name is None:
        name = filename

    if content is not None and not isinstance(content, (bytes, str)):
        content = json.dumps(content)

    if isinstance(content, str):
        content = content.encode()

    if isinstance(type, ArtifactType):
        type = type.value

    try:
        ArtifactType(type)
    except ValueError as exc:
        raise ValueError(f'ArtifactType enum has no value {type!r}.'
                         f'Supported types: {", ".join(ArtifactType.__members__)}') from exc

    create_event(
        'artifact',
        root_id=root_id,
        task_id=task_id,
        data={
            'name': str(name),
            'type': str(type),
            'category': str(category),
            'data': data if isinstance(data, dict) else {},
            'stream': str(stream) if stream else None,
            'filename': str(filename),
        },
        binary=content
    )


def create_artifact_from_traceback(
        name: Optional[str] = None,
        filename: Optional[str] = None,
        category: str = 'Intercepted',
        type: riberry.model.job.ArtifactType = riberry.model.job.ArtifactType.error,
):
    exc_type, exc, tb = sys.exc_info()

    if not exc_type:
        return

    if isinstance(exc, BaseError):
        error_content = f'{traceback.format_exc()}\n\n{"-" * 32}\n\n{json.dumps(exc.output(), indent=2)}'.encode()
    else:
        error_content = traceback.format_exc().encode()

    task = context.current.task
    create_artifact(
        name=name if name else f'Exception {task.name}',
        type=type,
        category=category,
        data={
            'Error Type': exc.__class__.__name__,
            'Error Message': str(exc),
        },
        filename=filename if filename else f'{task.name}-{task.request.id}.log',
        content=error_content
    )



import inspect


class Proxy:

    def __init__(self, getter):
        self.get = getter

    def __getattr__(self, item):
        return getattr(self.get(), item)

    def __repr__(self):
        return f'Proxy({self.get()})'


def function_path(func):
    return f'{inspect.getmodule(func).__name__}.{func.__name__}'


def internal_data_key(key):
    return f'_internal:{key}'



from riberry.celery import client
from riberry import model, policy


@policy.context.post_authorize(action='view')
def job_execution_by_id(execution_id):
    return model.job.JobExecution.query().filter_by(id=execution_id).one()


@policy.context.post_authorize(action='view')
def job_artifact_by_id(artifact_id):
    return model.job.JobExecutionArtifact.query().filter_by(id=artifact_id).one()


@policy.context.post_authorize(action='view')
def job_stream_by_id(stream_id):
    return model.job.JobExecutionStream.query().filter_by(id=stream_id).one()


@policy.context.post_authorize(action='view')
def delete_job_execution_by_id(execution_id):
    delete_job_execution(execution=job_execution_by_id(execution_id=execution_id))


@policy.context.post_authorize(action='view')
def delete_job_execution(execution):
    model.conn.delete(execution)


@policy.context.post_authorize(action='view')
def cancel_job_execution_by_id(execution_id):
    cancel_job_execution(execution=job_execution_by_id(execution_id=execution_id))


@policy.context.post_authorize(action='view')
def cancel_job_execution(execution):
    if execution.status in ('SUCCESS', 'FAILURE'):
        return

    user = policy.context.subject
    message = 'The current execution was manually cancelled by user {} ({}).'.format(
        user.username, user.details.display_name
    ).encode()

    artifact = model.job.JobExecutionArtifact(
        name=f'Workflow cancelled by user {user.username}',
        job_execution=execution,
        type=model.job.ArtifactType.error,
        category='Fatal',
        filename='fatal.log',
        size=len(message),
        binary=model.job.JobExecutionArtifactBinary(binary=message),
    )
    model.conn.add(artifact)
    client.workflow_complete(
        task_id=execution.task_id, root_id=execution.task_id, status='FAILURE', primary_stream=None)



import riberry
from ..env import current_context


def create_job(form_name, job_name=None, input_values=None, input_files=None, owner=None, execute=True):
    form: riberry.model.interface.Form = riberry.model.interface.Form.query().filter_by(
        internal_name=form_name,
    ).first()

    job_execution = current_context.current.job_execution
    job_execution_user = owner if owner else job_execution.creator

    with riberry.services.policy.policy_scope(user=job_execution_user):
        job = riberry.services.job.create_job(
            form_id=form.id,
            name=job_name or f'Via {job_execution.job.name} / #{job_execution.id}',
            input_values=input_values or {},
            input_files=input_files or {},
            execute=execute,
            parent_execution=job_execution,
        )

        riberry.model.conn.commit()
        return job



from typing import List

from celery.utils.log import logger
from sqlalchemy import desc, asc

from riberry import model
from riberry.celery import util, client


def start_tracking_execution(root_id):
    r = util.celery_redis_instance()
    instance = client.current_instance_name(raise_on_none=False)
    logger.info(f'tracker: Tracking workflow {root_id!r}')
    r.sadd(f'workflow:active:{instance}', root_id)


def check_stale_execution(app_instance):
    executions: List[model.job.JobExecution] = model.job.JobExecution.query().filter(
        model.job.JobExecution.status.in_(('ACTIVE', 'READY'))
    ).join(model.job.Job).order_by(
        desc(model.job.JobExecution.priority),
        asc(model.job.JobExecution.created)
    ).filter_by(instance=app_instance).all()

    if not executions:
        return

    r = util.celery_redis_instance()
    for execution in executions:
        if not r.sismember(f'workflow:active:{app_instance.internal_name}', execution.task_id):
            logger.info(f'tracker: Identified stale workflow {execution.task_id}')

            client.workflow_complete(
                task_id=execution.task_id,
                root_id=execution.task_id,
                status='FAILURE',
                primary_stream=None
            )

            client.wf.artifact(
                name='Workflow Cancelled',
                type=model.job.ArtifactType.error,
                category='Fatal',
                filename='fatal.log',
                content=str(
                    'The current workflow\'s ID was not found within Redis and has therefore been cancelled. This '
                    'usually occurs when Redis is flushed while an execution is active.'
                ).encode(),
                task_id=execution.task_id,
                root_id=execution.task_id
            )



from .base import AddonStartStopStep
import riberry
from riberry.app.backends.impl import celery as celery_backend


class ExternalTaskReceiver(riberry.app.addons.Addon):

    RECEIVER_QUEUE = 'rib.external'

    def register(self, riberry_app: 'riberry.app.base.RiberryApplication'):
        class ConcreteExternalTaskReceiverStep(ExternalTaskReceiverStep):
            rib = riberry_app

        riberry_app.backend.steps['worker'].add(ConcreteExternalTaskReceiverStep)
        riberry_app.backend.user_options['worker'].add(self.regiser_user_options)
        task_routes = {
            celery_backend.CeleryBackend.CHECK_EXTERNAL_TASK_NAME: {'queue': self.RECEIVER_QUEUE},
        }

        if not riberry_app.backend.conf.task_routes:
            riberry_app.backend.conf.task_routes = {}
        riberry_app.backend.conf.task_routes.update(task_routes)

        for addon in riberry_app.addons.values():
            if isinstance(addon, celery_backend.addons.Scale):
                addon.conf.ignore_queues.add(self.RECEIVER_QUEUE)

    @staticmethod
    def regiser_user_options(parser):
        parser.add_argument(
            '--rib-receiver', action='store_true', default=False,
            help='Receiver of external Riberry tasks.',
        )


class ExternalTaskReceiverStep(AddonStartStopStep):

    def __init__(self, worker, rib_receiver, **_):
        super().__init__(worker=worker, interval=1)
        self._is_receiver = bool(rib_receiver)

    def should_run(self) -> bool:
        return self._is_receiver

    def run(self):
        active = riberry.model.job.JobExecution.query().filter_by(
            status='ACTIVE'
        ).join(riberry.model.job.Job).filter_by(
            instance=self.rib.context.current.riberry_app_instance,
        ).join(riberry.model.job.JobExecutionExternalTask).filter_by(
            status='READY'
        ).count()

        operation = 'add' if active else 'cancel'

        consumer = self.worker.consumer
        queues = {q.name for q in consumer.task_consumer.queues}

        if operation == 'add' and ExternalTaskReceiver.RECEIVER_QUEUE not in queues:
            consumer.add_task_queue(ExternalTaskReceiver.RECEIVER_QUEUE)

        if operation == 'cancel' and ExternalTaskReceiver.RECEIVER_QUEUE in queues:
            consumer.cancel_task_queue(ExternalTaskReceiver.RECEIVER_QUEUE)



import traceback
import uuid

import pendulum

import riberry
from . import notify
from .. import current_riberry_app
from ..util import execution_tracker as tracker
from ..util.events import create_event


def queue_job_execution(execution: riberry.model.job.JobExecution):
    job = execution.job
    form = job.form

    try:
        execution.status = 'READY'
        execution.task_id = str(uuid.uuid4())
        tracker.start_tracking_execution(root_id=execution.task_id)
        riberry.model.conn.commit()

        execution_task_id = current_riberry_app.start(
            execution_id=execution.id,
            root_id=execution.task_id,
            form=form.internal_name,
        )

    except:
        execution.status = 'FAILURE'
        message = traceback.format_exc().encode()
        execution.artifacts.append(
            riberry.model.job.JobExecutionArtifact(
                job_execution=execution,
                name='Error on Startup',
                type='error',
                category='Fatal',
                filename='startup-error.log',
                size=len(message),
                binary=riberry.model.job.JobExecutionArtifactBinary(
                    binary=message
                )
            )
        )
        riberry.model.conn.commit()
        raise
    else:
        return execution_task_id


def execution_complete(task_id, root_id, status, stream):
    job: riberry.model.job.JobExecution = riberry.model.job.JobExecution.query().filter_by(
        task_id=root_id).first()
    if not job:
        return

    cxt = current_riberry_app.context
    with cxt.scope(root_id=root_id, task_id=root_id, stream=None, step=None, category=None):
        try:
            current_riberry_app.context.event_registry.call(
                event_type=current_riberry_app.context.event_registry.types.on_completion,
                status=status,
            )
        except:
            print('Error occurred while triggering on_completion event.')
            print(traceback.format_exc())
            riberry.model.conn.rollback()

    job.task_id = root_id
    job.status = status
    job.completed = job.updated = pendulum.DateTime.utcnow()
    if not job.started:
        job.started = pendulum.DateTime.utcnow()

    if stream is None:
        stream = riberry.model.job.JobExecutionStream.query().filter_by(task_id=root_id).first()
        if stream is not None:
            stream = stream.name

    if stream is not None:
        create_event(
            name='stream',
            root_id=root_id,
            task_id=root_id,
            data={
                'stream': stream,
                'state': status
            }
        )

    riberry.model.conn.commit()
    notify.workflow_complete(
        task_id=task_id,
        root_id=root_id,
        status=status,
    )


def execution_started(task, job_id, primary_stream):
    root_id = task.request.root_id

    job: riberry.model.job.JobExecution = riberry.model.job.JobExecution.query().filter_by(
        id=job_id,
    ).one()
    job.started = job.updated = pendulum.DateTime.utcnow()
    job.status = 'ACTIVE'
    job.task_id = root_id
    riberry.model.conn.commit()

    create_event(
        name='stream',
        root_id=root_id,
        task_id=root_id,
        data={
            'stream': primary_stream,
            'state': 'ACTIVE',
        }
    )



import pendulum
from sqlalchemy import Column, Sequence, Integer, MetaData
from sqlalchemy.ext.declarative import declarative_base

from riberry import model


class _BaseMixin:

    @classmethod
    def query(cls):
        return model.conn.query(cls)

    def __repr__(self):
        attribute_names = (['id'] if hasattr(self, 'id') else []) + getattr(self, '__reprattrs__', [])
        attributes = ', '.join(f'{attr}={repr(getattr(self, attr))}' for attr in attribute_names)
        return f'<{type(self).__name__} {attributes}>'


class _IdBuilder:

    def __init__(self, id_type):
        self.type = id_type

    def build(self, sequence='SEQUENCE_PK'):
        return Column(self.type, Sequence(sequence), primary_key=True)


meta = MetaData(
    naming_convention={
        'ix': 'ix_%(column_0_label)s',
        'uq': 'uq_%(table_name)s_%(column_0_N_name)s',
        'ck': 'ck_%(table_name)s_%(constraint_name)s',
        'fk': 'fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s',
        'pk': 'pk_%(table_name)s'
    }
)

Base = declarative_base(cls=_BaseMixin, metadata=meta)
id_builder = _IdBuilder(id_type=Integer)


def utc_now():
    return pendulum.DateTime.utcnow()



import json
import sys
import threading
import traceback
from contextlib import contextmanager

from io import BytesIO
from typing import Union, Optional, List
from celery import current_task, current_app

from riberry import model, config, policy, services
from riberry.celery.client.tasks import create_event, poll_external_task
from riberry.exc import BaseError
from riberry.model.job import ArtifactType


_cxt = threading.local()
_cxt.stream_name = None


class TaskWrap:

    def __init__(self, func, **kwargs):
        self.func = func
        self.kwargs = kwargs

    @property
    def name(self):
        return self.func.name

    def _mixin_kw(self, kwargs):
        return {**self.kwargs, **kwargs}

    def s(self, *args, **kwargs):
        return self.func.s(*args, **self._mixin_kw(kwargs=kwargs))

    def si(self, *args, **kwargs):
        return self.func.si(*args, **self._mixin_kw(kwargs=kwargs))

    def delay(self, *args, **kwargs):
        return self.func.delay(*args, **self._mixin_kw(kwargs=kwargs))


@contextmanager
def stream_context(stream: str):
    _cxt.stream_name = _validate_stream_name(stream)
    try:
        yield
    finally:
        _cxt.stream_name = None


def _validate_stream_name(stream: str):
    if not stream and not _cxt.stream_name:
        raise ValueError('Stream name cannot be blank')
    return str(stream or _cxt.stream_name)


def step(task, step: str = None, stream: str = None):
    step = step if step else task.name
    stream = _validate_stream_name(stream)
    return TaskWrap(task, __sb__=(stream, step))


def stream_start(task, stream: str = None):
    return TaskWrap(task, __ss__=_validate_stream_name(stream))


def stream_end(task, stream: str = None):
    return TaskWrap(task, __se__=_validate_stream_name(stream))


s = stream_start
e = stream_end
b = step


def artifact(filename: str, content: Union[bytes, str], name: str=None,
             type: Union[str, ArtifactType] = ArtifactType.output, category='Default', data: dict=None,
             stream: str=None, step=None, task_id: str=None, root_id: str=None):

    task_id = task_id or current_task.request.id
    root_id = root_id or current_task.request.root_id
    stream = stream or getattr(current_task, 'stream', None)
    step = step or getattr(current_task, 'step', None)

    if name is None:
        name = filename

    if isinstance(content, str):
        content = content.encode()

    if isinstance(type, ArtifactType):
        type = type.value

    try:
        ArtifactType(type)
    except ValueError as exc:
        raise ValueError(f'ArtifactType enum has no value {type!r}.'
                         f'Supported types: {", ".join(ArtifactType.__members__)}') from exc

    create_event(
        'artifact',
        root_id=root_id,
        task_id=task_id,
        data={
            'name': str(name),
            'type': str(type),
            'category': str(category),
            'data': data if isinstance(data, dict) else {},
            'stream': str(stream) if stream else None,
            'step': str(step) if step else None,
            'filename': str(filename),
        },
        binary=content
    )


def artifact_from_traceback(name=None, filename=None, category='Intercepted', type=model.job.ArtifactType.error):
    exc_type, exc, tb = sys.exc_info()

    if not exc_type:
        return

    if isinstance(exc, BaseError):
        error_content = f'{traceback.format_exc()}\n\n{"-"*32}\n\n{json.dumps(exc.output(), indent=2)}'.encode()
    else:
        error_content = traceback.format_exc().encode()

    artifact(
        name=name if name else f'Exception {current_task.name}',
        type=type,
        category=category,
        data={
            'Error Type': exc.__class__.__name__,
            'Error Message': str(exc)
        },
        filename=filename if filename else f'{current_task.name}-{current_task.request.id}.log',
        content=error_content
    )


def send_email(subject: str, body: str, mime_type: Optional[str]=None,
               sender: Optional[str] =None, receivers: Optional[List[str]]=None):

    if isinstance(receivers, str):
        receivers = [receivers]

    notify(
        notification_type='custom-email',
        data={
            'subject': subject,
            'mime_type': mime_type,
            'body': body,
            'from': sender,
            'to': receivers or []
        }
    )


def notify(notification_type, data=None, task_id=None, root_id=None):
    task_id = task_id or current_task.request.id
    root_id = root_id or current_task.request.root_id

    create_event(
        'notify',
        root_id=root_id,
        task_id=task_id,
        data={
            'type': notification_type,
            'data': data or {}
        },
        binary=None
    )


def current_execution() -> Optional[model.job.JobExecution]:
    return model.job.JobExecution.query().filter_by(
        task_id=current_task.request.root_id
    ).first()


def create_job(form_internal_name, job_name=None, input_values=None, input_files=None):
    form: model.interface.Form = model.interface.Form.query().filter_by(
        internal_name=form_internal_name,
    ).first()

    if not input_files:
        input_files = {}

    cleansed_input_files = {}
    for attr, value in input_files.items():
        if isinstance(value, str):
            value = value.encode()

        if isinstance(value, bytes):
            value = BytesIO(value)

        if not hasattr(value, 'read'):
            raise ValueError(f'wf.create_job:: value for input file {attr!r} must be of type str, bytes or stream')

        cleansed_input_files[attr] = value

    job_execution = current_execution()
    job_execution_user = job_execution.creator

    policy_provider = config.config.policies.provider
    with policy.context.scope(subject=job_execution_user, environment=None, policy_engine=policy_provider):
        job = services.job.create_job(
            form_id=form.id,
            name=job_name or f'Via {job_execution.job.name} / #{job_execution.id}',
            input_values=input_values or {},
            input_files=cleansed_input_files,
            execute=True,
            parent_execution=job_execution
        )

        model.conn.commit()
        return job


def create_external_task(name, task_type, external_task_id, input_data: bytes = None):
    external_task = model.job.JobExecutionExternalTask(
        job_execution=current_execution(),
        stream_id=None,
        task_id=external_task_id,
        name=name,
        type=task_type,
        input_data=input_data,
    )
    model.conn.add(external_task)
    model.conn.commit()
    return external_task


def poll_external_task_sig(external_task_id):
    return current_app.tasks['check-external-task'].si(external_task_id=external_task_id)





class BaseError(Exception):
    __msg__ = None
    __http_code__ = 500

    def __init__(self, target=None, data=None, **args):
        super(BaseError, self).__init__(self._fmt_message(**args, target=target))
        self.target = target
        self.exc_data = data or {}

    @classmethod
    def _fmt_message(cls, **args):
        return cls.__msg__.format(**args) if cls.__msg__ else f'{cls.__name__} was raised'

    def output(self):
        return {
            'code': type(self).__name__,
            'message': str(self),
            'target': self.target,
            'data': self.exc_data
        }


class AuthenticationError(BaseError):
    __msg__ = 'Wrong username or password supplied.'
    __http_code__ = 401

    def __init__(self):
        super(AuthenticationError, self).__init__(target='user')


class SessionExpired(BaseError):
    __msg__ = 'The user\'s session has expired.'
    __http_code__ = 401

    def __init__(self):
        super(SessionExpired, self).__init__(target='user')


class AuthorizationError(BaseError):
    __msg__ = 'User does not have access to the given resource.'
    __http_code__ = 403

    def __init__(self):
        super(AuthorizationError, self).__init__(target='user')


class ResourceNotFound(BaseError):
    __msg__ = 'The requested resource does not exist.'
    __http_code__ = 404

    def __init__(self, resource, identifier):
        super(ResourceNotFound, self).__init__(target=resource, data={
            'id': identifier
        })


class UnknownError(BaseError):
    __msg__ = 'An unknown error has occurred.'
    __http_code__ = 500

    def __init__(self, error=None):
        super(UnknownError, self).__init__()


class InputErrorGroup(BaseError):
    __msg__ = 'One or more errors occurred while validating the request.'
    __http_code__ = 400

    def __init__(self, *errors):
        super(InputErrorGroup, self).__init__(
            target=None,
            data={'errors': []}
        )
        self.extend(errors=errors)

    def extend(self, errors):
        self.exc_data['errors'] += [e.output() if isinstance(e, BaseError) else e for e in errors]


class RequiredInputError(BaseError):
    __msg__ = 'Required field {field!r} for {target} not provided.'
    __http_code__ = 400

    def __init__(self, target, field, internal_name=None):
        super(RequiredInputError, self).__init__(
            target=target,
            field=field,
            data=dict(internal=internal_name)
        )


class InvalidInputError(BaseError):
    __msg__ = 'Invalid {field!r} provided for {target}.'
    __http_code__ = 400

    def __init__(self, target, field, internal_name=None):
        super(InvalidInputError, self).__init__(
            target=target,
            field=field,
            data=dict(internal=internal_name)
        )


class InvalidEnumError(BaseError):
    __msg__ = 'Invalid {field!r} provided for {target}. Expected: {allowed_values}.'
    __http_code__ = 400

    def __init__(self, target, field, allowed_values, internal_name=None):
        super(InvalidEnumError, self).__init__(
            target=target,
            field=field,
            allowed_values=', '.join(repr(value) for value in allowed_values),
            data=dict(internal=internal_name)
        )


class UnknownInputError(BaseError):
    __msg__ = 'Unknown input field {field!r} provided for {target}.'
    __http_code__ = 400

    def __init__(self, target, field):
        super(UnknownInputError, self).__init__(target=target, field=field)


class UniqueInputConstraintError(BaseError):
    __msg__ = 'Cannot create {target} with {field}: {value!r}. This {field} is already in-use.'
    __http_code__ = 400

    def __init__(self, target, field, value):
        super(UniqueInputConstraintError, self).__init__(target=target, field=field, value=value)



import base64
import json

import pendulum
from celery import shared_task, current_app
from celery.utils.log import logger
from sqlalchemy import desc, asc

from riberry import model
from riberry.celery import client
from . import tracker


def _toggle_external_task_queue(app_instance):
    active = model.job.JobExecution.query().filter_by(
        status='ACTIVE'
    ).join(model.job.Job).filter_by(
        instance=app_instance,
    ).join(model.job.JobExecutionExternalTask).filter_by(
        status='READY'
    ).count()

    operation = 'add' if active else 'cancel'
    current_app.control.broadcast('toggle_external_task_queue', reply=False, arguments={'operation': operation})


@shared_task(ignore_result=True)
def poll():
    with model.conn:
        app_instance: model.application.ApplicationInstance = model.application.ApplicationInstance.query().filter_by(
            internal_name=client.current_instance_name(raise_on_none=True)
        ).first()
        if not app_instance:
            return

        tracker.check_stale_execution(app_instance=app_instance)
        _toggle_external_task_queue(app_instance=app_instance)
        if app_instance.status != 'online':
            return

        executions = model.job.JobExecution.query().filter(
            model.job.JobExecution.status == 'RECEIVED'
        ).join(model.job.Job).order_by(
            desc(model.job.JobExecution.priority),
            asc(model.job.JobExecution.created),
            asc(model.job.JobExecution.id),
        ).filter_by(instance=app_instance).all()

        for execution in executions:
            task = client.queue_job_execution(execution=execution)
            logger.info(f'poll - queueing task {task}')


@shared_task(ignore_result=True)
def echo():
    with model.conn:
        app_instance: model.application.ApplicationInstance = model.application.ApplicationInstance.query().filter_by(
            internal_name=client.current_instance_name(raise_on_none=True)
        ).first()
        if not app_instance:
            return

        heartbeat = model.application.Heartbeat.query().filter_by(instance=app_instance).first()
        if not heartbeat:
            heartbeat = model.application.Heartbeat(instance=app_instance)
            model.conn.add(heartbeat)

        heartbeat.updated = pendulum.DateTime.utcnow()
        model.conn.commit()


def create_event(name, root_id, task_id, data=None, binary=None):
    if not root_id:
        return

    if current_app.main != 'default':
        event_call = event.delay
    else:
        event_call = event

    event_call(
        name=name,
        time=pendulum.DateTime.utcnow().timestamp(),
        root_id=root_id,
        task_id=task_id,
        data=data,
        binary=base64.b64encode(binary).decode() if binary is not None else None
    )


@shared_task(ignore_result=True)
def event(name, time, task_id, root_id, data=None, binary=None):
    with model.conn:
        evt = model.misc.Event(
            name=name,
            time=time,
            task_id=task_id,
            root_id=root_id,
            data=json.dumps(data),
            binary=base64.b64decode(binary) if binary is not None else None
        )

        model.conn.add(evt)
        model.conn.commit()


@shared_task(queue='event', ignore_result=True)
def workflow_step_update(root_id, stream_name, step_name, task_id, status=None, note=None):
    with model.conn:
        job = model.job.JobExecution.query().filter_by(task_id=root_id).one()
        stream = model.job.JobExecutionStream().query().filter_by(job_execution=job, name=stream_name).first()
        step = model.job.JobExecutionStreamStep.query().filter_by(stream=stream, task_id=task_id).first()
        if not step:
            step = model.job.JobExecutionStreamStep(name=step_name, task_id=task_id)
            stream.steps.append(step)

        if status is not None:
            step.status = status
        if note is not None:
            step.note = note

        model.conn.commit()


@shared_task(queue='event', ignore_result=True)
def workflow_stream_update(root_id, stream_name, task_id, status):
    with model.conn:
        job = model.job.JobExecution.query().filter_by(task_id=root_id).one()
        stream = model.job.JobExecutionStream.query().filter_by(job_execution=job, name=stream_name).first()
        if not stream:
            stream = model.job.JobExecutionStream(name=stream_name, task_id=task_id)
            job.streams.append(stream)

        stream.status = status
        model.conn.commit()


@shared_task(bind=True, ignore_result=True)
def workflow_complete(task, status, primary_stream):
    with model.conn:
        return client.workflow_complete(task.request.id, task.request.root_id, status, primary_stream)


def poll_external_task(self, external_task_id):
    with model.conn:
        external_task = model.job.JobExecutionExternalTask.query().filter_by(
            task_id=external_task_id,
        ).first()

        if external_task:
            if external_task.status == 'WAITING':
                raise self.retry(countdown=1)
            elif external_task.status == 'READY':
                output_data = external_task.output_data
                if isinstance(output_data, bytes):
                    output_data = output_data.decode()
                external_task.status = 'COMPLETE'
                model.conn.commit()
                return output_data



import pendulum
from sqlalchemy import desc, asc

import riberry
from riberry.app.util import execution_tracker as tracker
from . import actions, env


def echo():
    with riberry.model.conn:
        app_instance = env.get_instance_model()

        heartbeat = riberry.model.application.Heartbeat.query().filter_by(instance=app_instance).first()
        if not heartbeat:
            heartbeat = riberry.model.application.Heartbeat(instance=app_instance)
            riberry.model.conn.add(heartbeat)

        heartbeat.updated = pendulum.DateTime.utcnow()
        riberry.model.conn.commit()


def poll():
    with riberry.model.conn:
        app_instance = env.get_instance_model()

        tracker.check_stale_execution(app_instance=app_instance)
        if app_instance.status != 'online':
            return

        executions = riberry.model.job.JobExecution.query().filter(
            riberry.model.job.JobExecution.status == 'RECEIVED'
        ).join(riberry.model.job.Job).order_by(
            desc(riberry.model.job.JobExecution.priority),
            asc(riberry.model.job.JobExecution.created),
            asc(riberry.model.job.JobExecution.id),
        ).filter_by(instance=app_instance).all()

        for execution in executions:
            execution_task_id = actions.executions.queue_job_execution(execution=execution)
            print(f'poll - queueing task {execution_task_id}')


def refresh():
    with riberry.model.conn:
        app_instance = env.get_instance_model()
        actions.shared_data.update_all_data_items(app_instance=app_instance)
        actions.reports.update_all_reports(app_instance=app_instance)



from celery import exceptions as celery_exc, current_task

import riberry
from riberry.app import actions
from riberry.app.misc.signals import task_prerun, task_postrun
from .extension import RiberryTask


class ExecutionComplete(Exception):
    pass


IGNORE_EXCEPTIONS = (
    celery_exc.Retry,
    celery_exc.SoftTimeLimitExceeded,
    celery_exc.TimeLimitExceeded,
)


def _retry_types(task_options):
    return tuple(list(IGNORE_EXCEPTIONS) + task_options.get('autoretry_for', []))


def _attempt_fallback(exc, task_options):
    fallback_on_error_provided, fallback_on_error = (
        'rib_fallback' in task_options, task_options.get('rib_fallback')
    )

    if not fallback_on_error_provided:
        actions.artifacts.create_artifact_from_traceback(category='Fatal')
        raise exc

    try:
        result = fallback_on_error() if callable(fallback_on_error) else fallback_on_error
        actions.artifacts.create_artifact_from_traceback(category='Intercepted')
        return result
    except:
        actions.artifacts.create_artifact_from_traceback(category='Fatal (intercept failed)')
        raise exc


class TaskExecutor:

    @property
    def riberry_app(self):
        return riberry.app.env.current_riberry_app

    def external_task_executor(self):
        def _external_task_executor(external_task_id, validator):
            external_task: riberry.model.job.JobExecutionExternalTask = riberry.model.job.JobExecutionExternalTask.query().filter_by(
                task_id=external_task_id,
            ).first()
            task: RiberryTask = self.riberry_app.context.current.task

            if external_task:
                if external_task.status == 'WAITING':
                    raise task.retry(countdown=1)
                elif external_task.status == 'READY':
                    output_data = external_task.output_data
                    if isinstance(output_data, bytes):
                        output_data = output_data.decode()

                    outcomes = self.riberry_app.context.event_registry.call(
                        event_type=self.riberry_app.context.event_registry.types.on_external_result_received,
                        key=validator,
                        kwargs=dict(
                            external_task=external_task,
                            result=output_data,
                        )
                    )

                    if outcomes:
                        assert len(outcomes) == 1, f'Multiple callbacks triggered for {validator}'
                        outcome = outcomes[0]
                        if outcome and outcome.retry:
                            external_task.status = 'WAITING'
                            external_task.input_data = outcome.input_data
                            riberry.model.conn.commit()
                            raise task.retry(countdown=1)

                    external_task.status = 'COMPLETE'
                    riberry.model.conn.commit()
                    return output_data

        return _external_task_executor

    def entry_point_executor(self):
        def _entry_point_executor(execution_id, form: str):
            entry_point = self.riberry_app.entry_points[form]
            actions.executions.execution_started(
                task=self.riberry_app.context.current.task,
                job_id=execution_id,
                primary_stream=entry_point.stream,
            )
            entry_point.func()

        return _entry_point_executor

    def riberry_task_executor(self, func, func_args, func_kwargs, task_options):
        riberry_properties = {}
        for key, value in list(func_kwargs.items()):
            if key.startswith('__rib_'):
                riberry_properties[key.replace('__rib_', '', 1)] = func_kwargs.pop(key)

        with riberry.model.conn:
            with self.riberry_app.context.scope(
                root_id=current_task.request.root_id,
                task_id=current_task.request.id,
                stream=riberry_properties.get('stream'),
                step=riberry_properties.get('step'),
                category=riberry_properties.get('category'),
            ):
                state = None
                mark_workflow_complete = False
                try:
                    task_prerun(context=self.riberry_app.context, props=riberry_properties)
                    result = self._execute_task(func, func_args, func_kwargs)
                    state = 'SUCCESS'
                    return result
                except ExecutionComplete:
                    state = 'FAILURE'
                    raise
                except celery_exc.Ignore:
                    state = 'IGNORED'
                    raise
                except Exception as exc:
                    state = 'FAILURE'
                    mark_workflow_complete = True
                    if isinstance(exc, _retry_types(task_options=task_options)) and not self._max_retries_reached(exc):
                        state = None
                        mark_workflow_complete = False
                        raise

                    result = _attempt_fallback(exc, task_options=task_options)
                    mark_workflow_complete = False
                    state = 'SUCCESS'
                    return result
                finally:
                    if mark_workflow_complete:
                        actions.executions.execution_complete(
                            task_id=self.riberry_app.context.current.task_id,
                            root_id=self.riberry_app.context.current.root_id,
                            status=state,
                            stream=None,
                        )
                    if state is not None:
                        task_postrun(context=self.riberry_app.context, props=riberry_properties, state=state)

    def _max_retries_reached(self, exc):
        active_task = self.riberry_app.context.current.task
        return bool(
            not isinstance(exc, celery_exc.Ignore) and
            active_task.max_retries is not None and
            active_task.request.retries >= active_task.max_retries
        )

    def _execute_task(self, func, args, kwargs):
        job_execution = self.riberry_app.context.current.job_execution
        if job_execution.status in ('FAILURE', 'SUCCESS'):
            raise ExecutionComplete(f'Execution {job_execution!r} is already marked as complete')
        return func(*args, **kwargs)

    def riberry_task_executor_wrapper(self, func, task_options):
        def wrapped_function(*args, **kwargs):
            return self.riberry_task_executor(
                func=func,
                func_args=args,
                func_kwargs=kwargs,
                task_options=task_options,
            )

        if 'name' not in task_options:
            task_options['name'] = riberry.app.util.misc.function_path(func=func)
        task_options['base'] = task_options.get('base') or RiberryTask

        return wrapped_function, task_options



from sqlalchemy.util.compat import contextmanager

from ..base import AddonStartStopStep
from .priority_queue import PriorityQueue
import riberry
from celery.utils.log import logger as log


class Capacity(riberry.app.addons.Addon):

    def __init__(self, parameter='capacity', key=None, sep='|', queue_cls=PriorityQueue, blocking: bool = True, block_retry: int = 0.5, r=None):
        self.parameter = parameter
        self.r = r or riberry.celery.util.celery_redis_instance()
        self.sep = sep
        self.queue = queue_cls(r=self.r, key=key, blocking=blocking, block_retry=block_retry)

    @property
    def last_value_key(self):
        return self.queue.make_key(self.queue.key, 'raw')

    @property
    def last_value(self):
        last_value = self.r.get(self.last_value_key)
        return last_value.decode() if isinstance(last_value, bytes) else None

    @last_value.setter
    def last_value(self, value):
        self.r.set(self.last_value_key, value='' if value is None else value)

    @contextmanager
    def borrow(self):
        member, score, version = self.queue.pop()
        try:
            yield member, score, version
        finally:
            self.queue.put(member=member, version=version)

    def register(self, riberry_app: 'riberry.app.base.RiberryApplication'):
        class ConcreteCapacityStep(CapacityStep):
            rib = riberry_app
            capacity = self

        if self.queue.key is None:
            self.queue.key = riberry_app.context.current.riberry_app_instance.internal_name

        riberry_app.backend.steps['worker'].add(ConcreteCapacityStep)


class CapacityStep(AddonStartStopStep):
    capacity: Capacity

    def __init__(self, worker, **_):
        super().__init__(worker=worker, interval=0.5)

    def should_run(self) -> bool:
        return True

    def run(self):
        value = self.rib.context.current.riberry_app_instance.active_schedule_value(name=self.capacity.parameter)

        if self.capacity.last_value is not None and value == self.capacity.last_value:
            return

        self.capacity.last_value = value
        values = [
            part.split(self.capacity.sep) if self.capacity.sep in part else (part, 0)
            for part in (value or '').split(' ')
        ]

        member_scores = {k: int(v) for k, v in values}
        self.capacity.queue.update(member_scores)
        log.info(
            f'DynamicPriorityParameter: ({self.capacity.queue.free_key}) updated {self.capacity.parameter} queue with {value!r}')



import time
import redis
import functools
from celery.utils.log import logger


class PriorityQueue:

    def __init__(
            self, r: redis.Redis, key: str, prefix: str='pq', sep: str=':', blocking: bool=True, block_retry: int=0.5):
        self.r: redis.Redis = r
        self.key = key
        self.prefix = prefix
        self.sep = sep
        self.blocking = blocking
        self.block_retry = block_retry

    def make_key(self, *args):
        sub_key = self.sep.join(map(str, args))
        return f'{self.prefix}{self.sep}{sub_key}'

    @property
    def version(self):
        return int(self.r.get(self.version_key) or 0)

    @version.setter
    def version(self, value):
        self.r.set(self.version_key, value=value)

    @property
    def version_key(self):
        return self.make_key(self.key, 'counter')

    @property
    def free_key(self):
        return self.generate_free_key(version=self.version)

    @property
    def lease_key(self):
        return self.generate_lease_key(version=self.version)

    def generate_free_key(self, version):
        return self.make_key(self.key, f'{version:09}', 'free')

    def generate_lease_key(self, version):
        return self.make_key(self.key, f'{version:09}', 'lease')

    def pop(self):
        while True:
            version = self.version
            result = self.r.transaction(
                functools.partial(self.pop_transaction, version=version),
                self.generate_free_key(version=version),
                self.generate_lease_key(version=version),
                value_from_callable=True
            )
            if not result and self.blocking:
                logger.warn(f'PriorityQueue: ({self.free_key}) encountered blank key, '
                            f'retrying after {self.block_retry} seconds.')
                time.sleep(self.block_retry)
                continue
            else:
                break

        return result, self.r.zscore(self.generate_free_key(version=version), result), version

    def pop_transaction(self, pipe: redis.client.Pipeline, version):
        free_key, lease_key = self.generate_free_key(version=version), self.generate_lease_key(version=version)
        [(member, score)] = pipe.zrevrange(free_key, 0, 0, withscores=True)
        member = member.decode()

        pipe.multi()
        pipe.zincrby(free_key, value=member, amount=-1)
        pipe.zincrby(lease_key, value=member, amount=1)
        return member

    def put(self, member, version):
        self.r.transaction(
            functools.partial(self.put_transaction, member=member, version=version),
            self.generate_free_key(version=version),
            self.generate_lease_key(version=version),
        )

    def put_transaction(self, pipe: redis.client.Pipeline, member, version):
        free_key, lease_key = self.generate_free_key(version=version), self.generate_lease_key(version=version)

        pipe.multi()
        pipe.zincrby(free_key, value=member, amount=1)
        pipe.zincrby(lease_key, value=member, amount=-1)

    def update(self, member_scores: dict):
        func = functools.partial(self.update_transaction, member_scores=member_scores)
        self.version = self.r.transaction(func, value_from_callable=True)

    def update_transaction(self, pipe: redis.client.Pipeline, member_scores):
        version = self.version + 1
        pipe.multi()
        pipe.zadd(self.generate_free_key(version=version), mapping=member_scores)
        return version

    def items(self):
        return [(k.decode(), v) for k, v in self.r.zrevrange(self.free_key, 0, -1, withscores=True)]

    def leased_items(self):
        return [(k.decode(), v) for k, v in self.r.zrevrange(self.lease_key, 0, -1, withscores=True)]

    def clear(self):
        self.r.delete(self.free_key, self.lease_key)

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.pop()
        except ValueError:
            raise StopIteration



from .engine import Rule, AttributeContext, Policy


class ShorthandRule(Rule):

    def __init__(self, func):
        self.func = func

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return self.func(context)


class ShorthandPolicy(Policy):

    def __init__(self, func, *collection):
        super(ShorthandPolicy, self).__init__(*collection)
        self.func = func

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return self.func(context)


class ShorthandPolicySet(Policy):

    def __init__(self, func, *collection):
        super(ShorthandPolicySet, self).__init__(*collection)
        self.func = func

    def target_clause(self, context: AttributeContext) -> bool:
        return True

    def condition(self, context: AttributeContext) -> bool:
        return self.func(context)


def rule(func):
    return ShorthandRule(func)


def policy(func):

    def builder(*collection):
        return ShorthandPolicy(func, *collection)

    return builder


def policy_set(func):

    def builder(*collection):
        return ShorthandPolicySet(func, *collection)

    return builder


import json

import click

from . import user, config_importer, groups


@click.group()
def cli():
    pass


@cli.command('import')
@click.option('--config-path', '-p', prompt='Configuration path', help='YAML file containing application references.')
@click.option('--commit/--rollback', '-c/-r', prompt='Dry run', is_flag=True,
              help='Gathers all database changes without commits.', default=False)
@click.option('--app', '-a', 'apps', help='Restricts the import to the specified app', multiple=True)
def importer(config_path, commit, apps):
    changes = config_importer.import_from_file(config_path, dry_run=not commit, restrict_apps=apps)
    print(json.dumps(changes, indent=2))


@cli.command('add-user')
@click.option('--username', prompt='Username', help="User's username")
@click.option('--password', prompt='Password', help="User's password", hide_input=True, confirmation_prompt=True)
@click.option('--first-name', prompt='First name', help="User's first name")
@click.option('--last-name', prompt='Last name', help="User's last name")
@click.option('--display-name', prompt='Full name', help="User's last name")
@click.option('--department', prompt='Department', help="User's department")
@click.option('--email', prompt='Email', help="User's email")
def add_user(username, password, first_name, last_name, display_name, department, email):
    user_id = user.add_user(
        username=username,
        password=password,
        first_name=first_name,
        last_name=last_name,
        display_name=display_name,
        department=department,
        email=email,
    )
    print(f'Created user {username} (User ID: {user_id})')


@cli.command('user-groups')
@click.argument('action', type=click.Choice(['add', 'remove']))
@click.option('--username', '-u', prompt='Username', help="User's username")
@click.option('--group', '-g', prompt='Username', help="Group's name")
def modify_user_groups(action, username, group):
    try:
        if action == 'add':
            groups.add_user_to_group(username=username, group_name=group)
            print(f'Added {username} to group {group}')
        elif action == 'remove':
            groups.remove_user_from_group(username=username, group_name=group)
            print(f'Removed {username} from group {group}')
    except Exception as exc:
        print(str(exc))
        exit(1)


if __name__ == '__main__':
    cli()



import json
from io import BytesIO
from typing import Dict

from riberry import model, services, policy, exc


class InputFileProxy(BytesIO):

    def __init__(self, obj, filename=None):
        self.filename = filename
        super().__init__(obj)

    @classmethod
    def from_object(cls, obj, filename='input'):
        if isinstance(obj, bytes):
            binary = obj
        elif isinstance(obj, str):
            binary = obj.encode()
            if not filename.endswith('.txt'):
                filename += '.txt'
        else:
            binary = json.dumps(obj).encode()
            if not filename.endswith('.json'):
                filename += '.json'

        return cls(binary, filename)


def jobs_by_form_id(form_id):
    return model.job.Job.query().filter_by(form_id=form_id).all()


def verify_inputs(input_value_definitions, input_file_definitions, input_values, input_files):
    value_map_definitions: Dict[str, 'model.interface.InputValueDefinition'] = {input_def.name: input_def for input_def in input_value_definitions}
    file_map_definitions: Dict[str, 'model.interface.InputValueDefinition'] = {input_def.name: input_def for input_def in input_file_definitions}

    value_mapping = {d.internal_name: d.name for d in value_map_definitions.values()}
    file_mapping = {d.internal_name: d.name for d in file_map_definitions.values()}

    input_values = {value_mapping.get(k, k): v for k, v in input_values.items()}
    input_files = {file_mapping.get(k, k): v for k, v in input_files.items()}

    input_value_mapping = {}
    input_file_mapping = {}
    errors = []

    for name, definition in value_map_definitions.items():
        if name in input_values:
            value = input_values.pop(name)
        else:
            value = definition.default_binary

        if definition.required and not value:
            err = exc.RequiredInputError(target='job', field=definition.name, internal_name=definition.internal_name)
            errors.append(err)
            continue

        if definition.allowed_binaries and value:
            values = value
            if isinstance(value, str):
                values = [value]

            if isinstance(values, list):
                values = [json.dumps(v).encode() if v else v for v in values]

            if set(values) - set(definition.allowed_binaries) or definition.type != 'text-multiple' and len(values) > 1:
                err = exc.InvalidEnumError(
                    target='job',
                    field=definition.name,
                    allowed_values=definition.allowed_values,
                    internal_name=definition.internal_name
                )
                errors.append(err)
                continue

        input_value_mapping[definition] = value

    for name in list(input_values):
        if name in file_mapping and name not in input_files:
            input_files[file_mapping[name]] = input_values.pop(name)

    for name, definition in file_map_definitions.items():
        if name in input_files:
            value = input_files.pop(name)
        else:
            value = None

        if definition.required and not value:
            err = exc.RequiredInputError(target='job', field=definition.name, internal_name=definition.internal_name)
            errors.append(err)
            continue

        input_file_mapping[definition] = value

    unexpected_inputs = set(input_values) | set(input_files)
    if unexpected_inputs:
        for input_ in unexpected_inputs:
            err = exc.UnknownInputError(target='job', field=input_)
            errors.append(err)

    if errors:
        raise exc.InputErrorGroup(*errors)

    return input_value_mapping, input_file_mapping


def create_job(form_id, name, input_values, input_files, execute, parent_execution=None):
    form = services.form.form_by_id(form_id=form_id)
    policy.context.authorize(form, action='view')

    errors = []
    if not name:
        err = exc.RequiredInputError(target='job', field='name')
        errors.append(err)
    else:
        if model.job.Job.query().filter_by(name=name).first():
            err = exc.UniqueInputConstraintError(target='job', field='name', value=name)
            errors.append(err)

    try:
        values_mapping, files_mapping = verify_inputs(
            input_value_definitions=form.input_value_definitions,
            input_file_definitions=form.input_file_definitions,
            input_values=input_values,
            input_files=input_files
        )
    except exc.InputErrorGroup as e:
        e.extend(errors)
        raise
    else:
        if errors:
            raise exc.InputErrorGroup(*errors)

    input_value_instances = []
    input_file_instances = []

    values_mapping = {
        k: (json.dumps(v).encode() if v and not isinstance(v, bytes) else v)
        for k, v in values_mapping.items()
    }

    for definition, value in values_mapping.items():
        input_value_instance = model.interface.InputValueInstance(
            name=definition.name,
            internal_name=definition.internal_name,
            raw_value=value
        )
        input_value_instances.append(input_value_instance)

    for definition, value in files_mapping.items():
        filename = definition.internal_name
        if not hasattr(value, 'read'):
            value = InputFileProxy.from_object(obj=value, filename=filename)

        binary = value.read()
        if isinstance(binary, str):
            binary = binary.encode()
        if hasattr(value, 'filename'):
            filename = value.filename

        input_file_instance = model.interface.InputFileInstance(
            name=definition.name,
            internal_name=definition.internal_name,
            filename=filename,
            binary=binary,
            size=len(binary) if binary else 0
        )
        input_file_instances.append(input_file_instance)

    job = model.job.Job(
        form=form,
        name=name,
        files=input_file_instances,
        values=input_value_instances,
        creator=policy.context.subject
    )

    policy.context.authorize(job, action='create')
    if execute:
        create_job_execution(job, parent_execution=parent_execution)

    model.conn.add(job)

    return job


@policy.context.post_authorize(action='view')
def job_by_id(job_id):
    return model.job.Job.query().filter_by(id=job_id).one()


@policy.context.post_filter(action='view')
def job_executions_by_id(job_id):
    return model.job.JobExecution.query().filter_by(job_id=job_id).all()


def create_job_execution_by_job_id(job_id):
    job = job_by_id(job_id=job_id)
    return create_job_execution(job=job)


def create_job_execution(job, parent_execution=None):
    execution = model.job.JobExecution(job=job, creator=policy.context.subject, parent_execution=parent_execution)

    policy.context.authorize(execution, action='create')
    model.conn.add(execution)

    return execution


def input_file_instance_by_id(input_file_instance_id) -> model.interface.InputFileInstance:
    return model.interface.InputFileInstance.query().filter_by(id=input_file_instance_id).one()


def delete_job_by_id(job_id):
    delete_job(job=job_by_id(job_id=job_id))


@policy.context.post_authorize(action='view')
def delete_job(job):
    model.conn.delete(job)



from contextlib import contextmanager

from riberry import model, config, policy


@contextmanager
def policy_scope(user=None, environment=None):
    if isinstance(user, str):
        user = model.auth.User.query().filter_by(username=user).one()

    policy_provider = config.config.policies.provider
    with policy.context.scope(subject=user, environment=environment, policy_engine=policy_provider):
        yield



from riberry import model


def add_user(username, password, first_name, last_name, display_name, department, email):
    user = model.auth.User(
        username=username,
        password=model.auth.User.secure_password(password).decode(),
        auth_provider='default',
        details=model.auth.UserDetails(
            first_name=first_name,
            last_name=last_name,
            display_name=display_name,
            department=department,
            email=email
        )
    )
    model.conn.add(user)
    model.conn.commit()

    return user.id



import riberry
import toml
import os
import binascii
import pathlib
import warnings

from riberry.util.common import variable_substitution

CONF_DEFAULT_BG_SCHED_INTERVAL = 10
CONF_DEFAULT_BG_EVENT_INTERVAL = 2
CONF_DEFAULT_BG_EVENT_PROCESS_LIMIT = 1000
CONF_DEFAULT_BG_CAPACITY_INTERVAL = 5

CONF_DEFAULT_DB_ECHO = False
CONF_DEFAULT_DB_CONN_PATH = (pathlib.Path(os.path.expanduser('~')) / '.riberry') / 'model.db'
CONF_DEFAULT_DB_CONN_URL = f'sqlite:///{CONF_DEFAULT_DB_CONN_PATH}'

CONF_DEFAULT_POLICY_PROVIDER = 'default'
CONF_DEFAULT_AUTH_PROVIDER = 'default'
CONF_DEFAULT_AUTH_TOKEN_PROVIDER = 'jwt'
CONF_DEFAULT_AUTH_TOKEN_PATH = (pathlib.Path(os.path.expanduser('~')) / '.riberry') / 'auth.key'
CONF_DEFAULT_AUTH_TOKEN_SIZE = 256


if 'RIBERRY_CONFIG_PATH' in os.environ:
    _config = variable_substitution(toml.load(os.environ['RIBERRY_CONFIG_PATH']))
else:
    warnings.warn(message=f'Environment variable \'RIBERRY_CONFIG_PATH\' not declared, '
                          f'defaulting to default configuration')
    _config = {}


def load_config_value(raw_config, default=None):
    if 'path' in raw_config:
        with open(raw_config['path']) as f:
            return f.read()
    elif 'envvar' in raw_config:
        return os.getenv(raw_config['envvar'])
    elif 'value' in raw_config:
        return raw_config['value']
    else:
        return default


class DatabaseConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        connection_config = self.raw_config.get('connection') or {}
        self.connection_url = load_config_value(connection_config)
        if not self.connection_url:
            CONF_DEFAULT_DB_CONN_PATH.parent.mkdir(exist_ok=True)
            self.connection_url = CONF_DEFAULT_DB_CONN_URL

        self.echo = connection_config.get('echo', CONF_DEFAULT_DB_ECHO)
        self.connection_arguments = self.raw_config.get('arguments', {})

    def enable(self):
        riberry.model.init(
            url=self.connection_url,
            echo=self.echo,
            connection_arguments=self.connection_arguments
        )


class AuthenticationTokenConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.provider = self.raw_config.get('provider') or CONF_DEFAULT_AUTH_TOKEN_PROVIDER

        self.secret = load_config_value(self.raw_config)
        if not self.secret:
            self.secret = self.make_secret()

    @staticmethod
    def make_secret():
        CONF_DEFAULT_AUTH_TOKEN_PATH.parent.mkdir(exist_ok=True)
        if not CONF_DEFAULT_AUTH_TOKEN_PATH.is_file():
            with open(CONF_DEFAULT_AUTH_TOKEN_PATH, 'wb') as f:
                f.write(binascii.hexlify(os.urandom(CONF_DEFAULT_AUTH_TOKEN_SIZE)))

        with open(CONF_DEFAULT_AUTH_TOKEN_PATH, 'rb') as f:
            return f.read().decode()


class AuthenticationProviderConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.default = self.raw_config.get('default') or CONF_DEFAULT_AUTH_PROVIDER
        self.supported = self.raw_config.get('supported') or [self.default]


class AuthenticationConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.providers = AuthenticationProviderConfig(self.raw_config.get('providers'))
        self.token = AuthenticationTokenConfig(self.raw_config.get('token'))
        self._config_cache = {}

    def __getitem__(self, item):
        if item not in self._config_cache:
            for provider in riberry.plugins.plugin_register['authentication']:
                if provider.name() == item:
                    self._config_cache[item] = provider(self.raw_config.get(item, {}))
                    break
            else:
                raise ValueError(f'Authentication provider {item!r} not found')

        return self._config_cache[item]

    @property
    def default_provider(self):
        return self[self.providers.default]

    def enable(self):
        for provider_name in self.providers.supported:
            self[provider_name].on_enabled()


class PolicyProviderConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.provider_name = self.raw_config.get('provider') or CONF_DEFAULT_POLICY_PROVIDER
        self._provider = None

    @property
    def provider(self):
        if self._provider is None:
            for provider in riberry.plugins.plugin_register['policies']:
                if provider.name == self.provider_name:
                    self._provider = provider
                    break
            else:
                raise ValueError(f'PolicyProviderConfig.provider:: '
                                 f'could not find register provider {self.provider_name!r}')
        return self._provider


class EmailNotificationConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self._enabled = config_dict.get('enabled', False)
        self.smtp_server = config_dict.get('smtpServer')
        self.sender = config_dict.get('sender')

    @property
    def enabled(self):
        return bool(self._enabled and self.smtp_server and self.sender)


class BackgroundTaskConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.events = BackgroundTaskEventsConfig(self.raw_config.get('events') or {})
        self.schedules = BackgroundTaskScheduleConfig(self.raw_config.get('schedules') or {})
        self.capacity = BackgroundTaskCapacityConfig(self.raw_config.get('capacity') or {})


class BackgroundTaskEventsConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.interval = config_dict.get('interval', CONF_DEFAULT_BG_EVENT_INTERVAL)
        self.processing_limit = config_dict.get('limit', CONF_DEFAULT_BG_EVENT_PROCESS_LIMIT)


class BackgroundTaskScheduleConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.interval = config_dict.get('interval', CONF_DEFAULT_BG_SCHED_INTERVAL)


class BackgroundTaskCapacityConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict or {}
        self.interval = config_dict.get('interval', CONF_DEFAULT_BG_CAPACITY_INTERVAL)


class RiberryConfig:

    def __init__(self, config_dict):
        self.raw_config = config_dict
        self.authentication = AuthenticationConfig(self.raw_config.get('authentication') or {})
        self.policies = PolicyProviderConfig(self.raw_config.get('policies') or {})
        self.database = DatabaseConfig(self.raw_config.get('database') or {})
        if 'notification' in self.raw_config and isinstance(self.raw_config['notification'], dict):
            email_config = self.raw_config['notification'].get('email') or {}
        else:
            email_config = {}

        self.email = EmailNotificationConfig(email_config)
        self.background = BackgroundTaskConfig(self.raw_config.get('background') or {})

    @property
    def celery(self):
        return self.raw_config.get('celery') or {}

    def enable(self):
        self.authentication.enable()
        self.database.enable()


config: RiberryConfig = RiberryConfig(config_dict=_config)



import base64
import functools
import os
import traceback
import uuid
import warnings
from typing import Dict, Tuple

import pendulum
from celery import Celery, bootsteps
from celery import current_task
from celery import exceptions as celery_exc
from celery.result import AsyncResult

from riberry import model
from riberry.celery.client import tasks
from riberry.celery.client.dynamic import DynamicParameters
from . import wf, signals, scale, dynamic, tracker, control


warnings.warn(
    'The module riberry.celery.client is deprecated. Please use riberry.celery.app instead',
    DeprecationWarning,
    stacklevel=2
)


IGNORE_EXCEPTIONS = (
    celery_exc.Ignore,
    celery_exc.Retry,
    celery_exc.SoftTimeLimitExceeded,
    celery_exc.TimeLimitExceeded
)

BYPASS_ARGS = (
    '__ss__', '__se__', '__sb__'
)


def current_instance_name(raise_on_none=False) -> str:
    name = os.getenv('RIBERRY_INSTANCE')
    if name is None and raise_on_none:
        raise EnvironmentError("Environment variable 'RIBERRY_INSTANCE' not set")
    return name


def is_current_instance(instance_name: str) -> bool:
    return bool(instance_name) and current_instance_name(raise_on_none=False) == instance_name


def queue_job_execution(execution: model.job.JobExecution):

    job = execution.job
    form = job.form
    app_instance = job.instance

    application_name = app_instance.application.internal_name
    workflow_app = Workflow.by_internal_name(internal_name=application_name)

    try:
        execution.status = 'READY'
        execution.task_id = str(uuid.uuid4())
        tracker.start_tracking_execution(root_id=execution.task_id)
        model.conn.commit()

        task = workflow_app.start(
            execution_id=execution.id,
            root_id=execution.task_id,
            form=form.internal_name,
            input_values={v.internal_name: v.value for v in job.values},
            input_files={v.internal_name: base64.b64encode(v.binary).decode() for v in job.files}
        )
    except:
        execution.status = 'FAILURE'
        message = traceback.format_exc().encode()
        execution.artifacts.append(
            model.job.JobExecutionArtifact(
                job_execution=execution,
                name='Error on Startup',
                type='error',
                category='Fatal',
                filename='startup-error.log',
                size=len(message),
                binary=model.job.JobExecutionArtifactBinary(
                    binary=message
                )
            )
        )
        model.conn.commit()
        raise
    else:
        return task


def workflow_complete(task_id, root_id, status, primary_stream):

    job: model.job.JobExecution = model.job.JobExecution.query().filter_by(task_id=root_id).first()
    if not job:
        return
    
    job.task_id = root_id
    job.status = status
    job.completed = job.updated = pendulum.DateTime.utcnow()
    if not job.started:
        job.started = pendulum.DateTime.utcnow()

    if primary_stream is None:
        stream = model.job.JobExecutionStream.query().filter_by(task_id=root_id).first()
        if stream is not None:
            primary_stream = stream.name

    if primary_stream is not None:
        tasks.create_event(
            name='stream',
            root_id=root_id,
            task_id=root_id,
            data={
                'stream': primary_stream,
                'state': status
            }
        )

    model.conn.commit()
    wf.notify(
        notification_type='workflow_complete',
        data=dict(status=status),
        task_id=task_id,
        root_id=root_id
    )


def is_workflow_complete(task):
    root_id = task.request.root_id
    job: model.job.JobExecution = model.job.JobExecution.query().filter_by(task_id=root_id).first()
    return job.status in ('FAILURE', 'SUCCESS') if job else True


def workflow_started(task, job_id, primary_stream):
    root_id = task.request.root_id

    job: model.job.JobExecution = model.job.JobExecution.query().filter_by(id=job_id).one()
    job.started = job.updated = pendulum.DateTime.utcnow()
    job.status = 'ACTIVE'
    job.task_id = root_id
    task.stream = primary_stream
    model.conn.commit()

    tasks.create_event(
        name='stream',
        root_id=root_id,
        task_id=root_id,
        data={
            'stream': primary_stream,
            'state': 'ACTIVE'
        }
    )

    wf.notify(notification_type='workflow_started')


def execute_task(func, func_args, func_kwargs, task_kwargs):
    # noinspection PyBroadException
    try:
        return func(*func_args, **func_kwargs)
    except tuple(list(IGNORE_EXCEPTIONS) + task_kwargs.get('autoretry_for', [])):
        raise
    except Exception as exc:
        wf.artifact_from_traceback(category='Intercepted' if 'rib_fallback' in task_kwargs else 'Fatal')

        if 'rib_fallback' in task_kwargs:
            fallback = task_kwargs.get('rib_fallback')
            return fallback() if callable(fallback) else fallback
        else:
            workflow_complete(current_task.request.id, current_task.request.root_id, status='FAILURE', primary_stream=None)
            raise


def bypass(func, **task_kwargs):
    @functools.wraps(func)
    def inner(*args, **kwargs):

        if not task_kwargs.get('rib_task', True):
            return func(*args, **kwargs)

        with model.conn:
            if is_workflow_complete(current_task):
                AsyncResult(current_task.request.id).revoke()
                raise Exception('Workflow already cancelled')

            filtered_kwargs = {k: v for k, v in kwargs.items() if k not in BYPASS_ARGS}
            return execute_task(
                func=func,
                func_args=args,
                func_kwargs=filtered_kwargs,
                task_kwargs=task_kwargs
            )

    return inner


def patch_task(task):

    def stream_start(stream: str = None):
        return wf.s(task, stream=stream)

    def stream_end(stream: str = None):
        return wf.e(task, stream=stream)

    def step(step: str = None, stream: str = None):
        return wf.b(task, step=step, stream=stream)

    task.stream_start = stream_start
    task.stream_end = stream_end
    task.step = step

    return task


def patch_app(app):
    def task_deco(*args, **kwargs):
        if len(args) == 1:
            if callable(args[0]):
                return patch_task(Celery.task(app, bypass(*args, **kwargs), **kwargs))
            raise TypeError('argument 1 to @task() must be a callable')

        def inner(func):
            return patch_task(Celery.task(app, **kwargs)(bypass(func, **kwargs)))

        return inner

    app.task = task_deco
    app.task(name='check-external-task', bind=True, max_retries=None)(wf.poll_external_task)
    return app


class WorkflowEntry:

    def __init__(self, name, func, primary_stream, primary_step):
        self.name = name
        self.func = func
        self.primary_stream = primary_stream
        self.primary_step = primary_step


class Workflow:
    __registered__ = {}

    def __init__(self, name, app, beat_queue=None, event_queue=None, scalable_queues=None, dynamic_parameters=None):
        self.name = name
        self.beat_queue = beat_queue or 'rib.beat'
        self.event_queue = event_queue or 'rib.event'
        self.__registered__[name] = self
        self.app = patch_app(app)
        self.scale = scale.ConcurrencyScale(self.app, target_queues=scalable_queues) if scalable_queues else None
        self.form_entries: Dict[Tuple[str, str], WorkflowEntry] = {}
        self.entry_point = self._make_entry_point(self.app, self.form_entries)
        self._extend_cli(app)
        self._configure_beat_queues(app, self.beat_queue)
        self._configure_event_queue(app, self.event_queue)
        self._configure_manual_task_queue(app)
        self.dynamic_parameters = DynamicParameters(
            riberry_workflow=self,
            handlers=dynamic_parameters,
            beat_queue=self.beat_queue
        )

    @classmethod
    def by_internal_name(cls, internal_name):
        return Workflow.__registered__[internal_name]

    @staticmethod
    def _configure_beat_queues(app, beat_queue):
        schedule = {
            'poll-executions': {
                'task': 'riberry.celery.client.tasks.poll',
                'schedule': 2,
                'options': {'queue': beat_queue}
            },
            'echo-status': {
                'task': 'riberry.celery.client.tasks.echo',
                'schedule': 2,
                'options': {'queue': beat_queue}
            }
        }

        if not app.conf.beat_schedule:
            app.conf.beat_schedule = {}
        app.conf.beat_schedule.update(schedule)

    @staticmethod
    def _configure_event_queue(app, event_queue):
        task_routes = {
            'riberry.celery.client.tasks.event': {'queue': event_queue},
        }

        if not app.conf.task_routes:
            app.conf.task_routes = {}
        app.conf.task_routes.update(task_routes)

    @staticmethod
    def _configure_manual_task_queue(app):
        task_routes = {
            'check-external-task': {'queue': 'rib.manual'},
        }

        if not app.conf.task_routes:
            app.conf.task_routes = {}
        app.conf.task_routes.update(task_routes)

    @staticmethod
    def _extend_cli(app):

        def rib_instance_cli(parser):
            parser.add_argument('--rib-instance', default=None, help='Defines the Riberry instance')

        class RiberryInstanceStep(bootsteps.StartStopStep):

            def __init__(self, worker, rib_instance=None, **options):
                super(RiberryInstanceStep, self).__init__(worker, **options)
                self.rib_instance = rib_instance or os.getenv('RIBERRY_INSTANCE')

            def start(self, worker):
                os.environ['RIBERRY_INSTANCE'] = self.rib_instance

        if not os.environ.get('RIBERRY_TESTSUITE'):
            app.user_options['worker'].add(rib_instance_cli)
            app.steps['worker'].add(RiberryInstanceStep)

    @staticmethod
    def _make_entry_point(app, form_entries: Dict[Tuple[str, str], WorkflowEntry]):
        @app.task(bind=True)
        def entry_point(task, execution_id, name: str, values: Dict, files: Dict):
            with model.conn:
                workflow_entry = form_entries[name]
                workflow_started(task, execution_id, workflow_entry.primary_stream)
                workflow_entry.func(task, **values, **files)

        return entry_point

    def entry(self, name, primary_stream='Overall', primary_step='Entry'):
        def wrapper(func):
            self.form_entries[name] = WorkflowEntry(
                name=name,
                func=func,
                primary_stream=primary_stream,
                primary_step=primary_step,
            )

        return wrapper

    def start(self, execution_id, root_id, input_name, input_values, input_files):
        if input_name not in self.form_entries:
            raise ValueError(f'Application {self.name!r} does not have an entry point with '
                             f'name {input_name!r} registered.')

        workflow_entry: WorkflowEntry = self.form_entries[input_name]

        with wf.stream_context(stream=workflow_entry.primary_stream):
            body = wf.b(self.entry_point, step=workflow_entry.primary_step).si(
                execution_id=execution_id,
                name=input_name,
                values=input_values,
                files=input_files,
            )

        callback_success = tasks.workflow_complete.si(status='SUCCESS', primary_stream=workflow_entry.primary_stream)
        callback_failure = tasks.workflow_complete.si(status='FAILURE', primary_stream=workflow_entry.primary_stream)

        body.options['root_id'] = root_id
        callback_success.options['root_id'] = root_id
        callback_failure.options['root_id'] = root_id

        task = body.on_error(callback_failure) | callback_success

        task.options['root_id'] = root_id

        return task.apply_async()



import datetime
from collections import Mapping, defaultdict
from contextlib import contextmanager
from operator import itemgetter
from typing import AnyStr, Any, Dict, Iterator

import pendulum
import time
from sqlalchemy.exc import IntegrityError

import riberry


class SharedExecutionData(Mapping):

    def __init__(self, context):
        self.context: riberry.app.context.Context = context
        self._lock: Dict[AnyStr, riberry.model.misc.ResourceData] = {}
        self._dirty = set()
        self._listeners = defaultdict(list)

    def listen(self, key, callback):
        self._listeners[key].append(callback)

    def __getitem__(self, item: AnyStr) -> Any:
        return self._get_instance(key=item).value

    def __setitem__(self, key, value):
        if key not in self._lock:
            raise PermissionError(f'No lock present for data key {key!r}')

        if pendulum.DateTime.utcnow() > pendulum.instance(self._lock[key].expiry):
            raise TimeoutError(f'Lock for data key {key!r} has expired!')

        self._lock[key].value = value
        self._dirty.add(key)

    def __delitem__(self, key):
        instance = self._get_instance(key=key)
        riberry.model.conn.delete(instance=instance)
        riberry.model.conn.commit()

    def __len__(self) -> int:
        return riberry.model.misc.ResourceData.query().filter_by(
            resource_id=self.context.current.job_execution.id,
            resource_type=riberry.model.misc.ResourceType.job_execution,
        ).count()

    def __iter__(self) -> Iterator[AnyStr]:
        instances = riberry.model.conn.query(
            riberry.model.misc.ResourceData.name
        ).filter_by(
            resource_id=self.context.current.job_execution.id,
            resource_type=riberry.model.misc.ResourceType.job_execution,
        ).all()

        return map(itemgetter(0), instances)

    def _get_instance(self, key):
        job_execution = self.context.current.job_execution

        # check if key exists
        instance = riberry.model.misc.ResourceData.query().filter_by(
            resource_id=job_execution.id,
            resource_type=riberry.model.misc.ResourceType.job_execution,
            name=key,
        ).first()

        # create if it doesn't
        if not instance:
            instance = riberry.model.misc.ResourceData(
                resource_id=job_execution.id,
                resource_type=riberry.model.misc.ResourceType.job_execution,
                name=key
            )
            try:
                riberry.model.conn.add(instance)
                riberry.model.conn.commit()
            except IntegrityError:
                riberry.model.conn.rollback()
                return self._get_instance(key=key)

        return instance

    def _acquire_lock(self, key, ttl, poll_interval):
        instance = self._get_instance(key=key)
        lock_value = self.context.current.task_id

        while True:
            riberry.model.conn.query(riberry.model.misc.ResourceData).filter(
                (riberry.model.misc.ResourceData.id == instance.id) &
                (
                    (riberry.model.misc.ResourceData.lock == None) |
                    (riberry.model.misc.ResourceData.expiry < datetime.datetime.now(tz=datetime.timezone.utc))
                )
            ).update({
                'lock': lock_value,
                'expiry': pendulum.DateTime.utcnow().add(seconds=ttl)
            })
            riberry.model.conn.commit()
            riberry.model.conn.expire(instance=instance)

            if instance.lock != lock_value:
                time.sleep(poll_interval)
            else:
                self._lock[key] = instance
                break

    def _release_lock(self, key):
        if self._lock:
            instance = self._lock.pop(key)
            instance.lock = None
            instance.expiry = None
            instance.marked_for_refresh = instance.marked_for_refresh or (key in self._dirty)
            riberry.model.conn.commit()

            if key in self._dirty:
                for listener in self._listeners[key]:
                    listener(key)
                self._dirty.remove(key)

    @contextmanager
    def lock(self, key, ttl=60, poll_interval=1):
        try:
            yield self._acquire_lock(key=key, ttl=ttl, poll_interval=poll_interval)
        finally:
            self._release_lock(key=key)

    def execute_once(self, key, func):
        with self.context.data.lock(key=key):
            if not self.context.data[key]:
                self.context.data[key] = True
                func()

    def set(self, key, value, **kwargs):
        with self.lock(key=key, **kwargs):
            self[key] = value() if callable(value) else value



import time

from celery import bootsteps
import riberry
from celery.utils.log import logger as log


class AddonStartStopStep(bootsteps.StartStopStep):
    requires = {'celery.worker.components:Timer'}

    rib: 'riberry.app.base.RiberryApplication'

    def __init__(self, worker, interval, priority=10, **kwargs):
        super().__init__(worker, interval=interval, priority=priority, **kwargs)

        self._timer = None
        self.interval = interval
        self.priority = priority
        self._worker = worker

    @property
    def worker(self):
        return self._worker

    @property
    def worker_state(self):
        return self.worker.state

    @property
    def consumer(self):
        return self.worker.consumer

    def start(self, worker):
        if self.should_run():
            self._timer = worker.timer.call_repeatedly(
                self.interval, self.__execute, priority=self.priority,
            )

    def stop(self, worker):
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None

    def __execute(self):
        start_time = time.time()
        try:
            log.debug(f'Started {self.step_name}')
            riberry.model.conn.raw_engine.dispose()
            with riberry.model.conn:
                self.run()
        finally:
            log.debug(f'Completed {self.step_name} in {time.time() - start_time:2} seconds')

    def should_run(self) -> bool:
        raise NotImplementedError

    def run(self):
        raise NotImplementedError

    @property
    def step_name(self):
        return type(self).__name__



from riberry import policy as rib_policy
from . import application, application_instance, form, auth, job, job_executions, self, policy


def fetch_relationship(model_object, attribute, action):
    resource = getattr(model_object, attribute)

    if isinstance(resource, list):
        return rib_policy.context.filter(resources=resource, action=action)
    if rib_policy.context.authorize(resource=resource, action=action, on_deny=None) is not False:
        return resource
    return None



import riberry
from typing import List
from .. import env


def update_all_data_items(app_instance=None):
    app_instance = app_instance if app_instance else env.get_instance_model()
    data_items: List[riberry.model.misc.ResourceData] = riberry.model.misc.ResourceData.query().filter(
        riberry.model.misc.ResourceData.marked_for_refresh == True,
        riberry.model.misc.ResourceData.resource_type == riberry.model.misc.ResourceType.job_execution,
    ).join(
        riberry.model.job.JobExecution,
        riberry.model.misc.ResourceData.resource_id == riberry.model.job.JobExecution.id,
    ).join(
        riberry.model.job.Job
    ).filter_by(
        instance=app_instance
    ).all()

    cxt = env.current_context
    for item in data_items:
        root_id = riberry.model.job.JobExecution.query().filter_by(id=item.resource_id).one().task_id

        with cxt.scope(root_id=root_id, task_id=root_id, stream=None, step=None, category=None):
            cxt.event_registry.call(event_type=cxt.event_registry.types.on_data_updated, data_name=item.name)
            cxt.event_registry.call(event_type=cxt.event_registry.types.on_report_refresh, data_name=item.name)
            if item.marked_for_refresh:
                item.marked_for_refresh = False
                riberry.model.conn.commit()



from typing import Tuple

from riberry import model, services
from sqlalchemy.orm.exc import NoResultFound


def _find_user_and_group(username: str, group_name: str) -> Tuple[model.auth.User, model.group.Group]:
    try:
        user = model.auth.User.query().filter_by(username=username).one()
    except NoResultFound:
        raise ValueError(f'Could not find user with username {username!r}')

    try:
        group = model.group.Group.query().filter_by(name=group_name).one()
    except NoResultFound:
        raise ValueError(f'Could not find group with name {group_name!r}')

    return user, group


def add_user_to_group(username: str, group_name: str):
    user, group = _find_user_and_group(username=username, group_name=group_name)
    services.auth.add_user_to_group(group_id=group.id, user_id=user.id)
    model.conn.commit()


def remove_user_from_group(username: str, group_name: str):
    user, group = _find_user_and_group(username=username, group_name=group_name)
    services.auth.remove_user_from_group(group_id=group.id, user_id=user.id)
    model.conn.commit()



from typing import List, Dict

from riberry import model, policy, services


@policy.context.post_filter(action='view')
def all_application_instances() -> List[model.application.ApplicationInstance]:
    return model.application.ApplicationInstance.query().all()


@policy.context.post_authorize(action='view')
def application_instance_by_id(application_instance_id) -> model.application.ApplicationInstance:
    return model.application.ApplicationInstance.query().filter_by(id=application_instance_id).one()


@policy.context.post_authorize(action='view')
def application_instance_by_internal_name(internal_name) -> model.application.ApplicationInstance:
    return model.application.ApplicationInstance.query().filter_by(internal_name=internal_name).one()


@policy.context.post_filter(action='view')
def instances_by_application_id(application_id) -> List[model.application.ApplicationInstance]:
    application = services.application.application_by_id(application_id=application_id)
    return application.instances


def create_application_instance(application, name, internal_name, schedules: List[Dict]) -> model.application.ApplicationInstance:
    application_instance = model.application.ApplicationInstance(
        application=application,
        name=name,
        internal_name=internal_name,
        schedules=create_application_instance_schedules(attributes_dict=schedules),
    )

    policy.context.authorize(application_instance, action='create')
    model.conn.add(application_instance)
    return application_instance


def create_application_instance_schedules(attributes_dict):
    return [
        model.application.ApplicationInstanceSchedule(
            days=schedule['days'],
            start_time=schedule['start_time'],
            end_time=schedule['end_time'],
            timezone=schedule['timezone'],
            parameter=schedule['parameter'],
            value=schedule['value'],
            priority=schedule['priority'],
        ) for schedule in attributes_dict
    ]


def update_application_instance(application_instance: model.application.ApplicationInstance, attributes: Dict):
    for attr in {'name'} & set(attributes):
        setattr(application_instance, attr, attributes[attr])
    return application_instance



from typing import Dict, AnyStr

import celery

import riberry
from . import patch, tasks, addons
from .executor import TaskExecutor


def send_task_process_rib_kwargs(self, *args, **kwargs):
    riberry_properties = {}
    if kwargs.get('kwargs'):
        for key, value in kwargs['kwargs'].items():
            if key.startswith('__rib_'):
                riberry_properties[key.replace('__rib_', '', 1)] = value


class CeleryBackend(riberry.app.backends.RiberryApplicationBackend):
    instance: celery.Celery

    ENTRY_POINT_TASK_NAME = 'riberry.core.app.entry_point'
    CHECK_EXTERNAL_TASK_NAME = 'riberry.core.app.check_external_task'

    def __init__(self, instance):
        super().__init__(instance=instance)
        self.executor = TaskExecutor()

    def initialize(self):
        patch.patch_send_task(instance=self.instance, func=send_task_process_rib_kwargs)

        # Register "entry point" task
        self.task(
            name=self.ENTRY_POINT_TASK_NAME,
        )(self.executor.entry_point_executor())

        # Register "external task checker" task
        self.task(
            name=self.CHECK_EXTERNAL_TASK_NAME,
            max_retries=None,
        )(self.executor.external_task_executor())

    def default_addons(self) -> Dict[AnyStr, 'riberry.app.addons.Addon']:
        return {
            'scale': addons.Scale(),
            'background': addons.BackgroundTasks(),
            'external-receiver': addons.ExternalTaskReceiver(),
        }

    def register_task(self, func, **options) -> celery.Task:
        wrapped_func, options = self.executor.riberry_task_executor_wrapper(func=func, task_options=options)
        return self.instance.task(**options)(wrapped_func)

    def task_by_name(self, name: AnyStr):
        return self.instance.tasks[name]

    def start_execution(self, execution_id, root_id, entry_point) -> AnyStr:
        task = self.task_by_name(self.ENTRY_POINT_TASK_NAME)
        task_signature = task.si(execution_id=execution_id, form=entry_point.form)

        callback_success = tasks.execution_complete.si(status='SUCCESS', stream=entry_point.stream)
        callback_failure = tasks.execution_complete.si(status='FAILURE', stream=entry_point.stream)

        task_signature.options['root_id'] = root_id
        callback_success.options['root_id'] = root_id
        callback_failure.options['root_id'] = root_id

        exec_signature = task_signature.on_error(callback_failure) | callback_success
        exec_signature.options['root_id'] = root_id

        riberry.app.util.events.create_event(
            name='stream',
            root_id=root_id,
            task_id=root_id,
            data={
                'stream': entry_point.stream,
                'state': 'QUEUED',
            }
        )

        return exec_signature.apply_async().id

    def create_receiver_task(self, external_task_id, validator):
        return self.task_by_name(self.CHECK_EXTERNAL_TASK_NAME).si(
            external_task_id=external_task_id,
            validator=validator,
        )

    def active_task(self):
        return celery.current_task



from typing import List, Dict

from sqlalchemy import desc

from riberry import model, policy
from riberry import services


@policy.context.post_filter(action='view')
def all_forms() -> List[model.interface.Form]:
    return model.interface.Form.query().all()


@policy.context.post_authorize(action='view')
def form_by_id(form_id) -> model.interface.Form:
    return model.interface.Form.query().filter_by(id=form_id).one()


@policy.context.post_authorize(action='view')
def form_by_internal_name(internal_name) -> model.interface.Form:
    return model.interface.Form.query().filter_by(
        internal_name=internal_name,
    ).one()


@policy.context.post_filter(action='view')
def forms_by_application_id(application_id):
    application = services.application.application_by_id(application_id=application_id)
    return application.forms


def create_form(application, instance, name, internal_name, version, description, input_files, input_values) -> model.interface.Form:
    form = model.interface.Form(
        application=application,
        instance=instance,
        name=name,
        internal_name=internal_name,
        version=version,
        description=description,
        input_file_definitions=[model.interface.InputFileDefinition(**d) for d in input_files],
        input_value_definitions=[model.interface.InputValueDefinition(**d) for d in input_values]
    )

    policy.context.authorize(form, action='create')
    model.conn.add(form)
    return form


def update_form(form: model.interface.Form, attributes: Dict):
    for attr in {'name', 'version', 'description'} & set(attributes):
        setattr(form, attr, attributes[attr])
    return form


@policy.context.post_authorize(action='view')
def file_definition_by_internal_name(form, internal_name) -> model.interface.InputFileDefinition:
    return model.interface.InputFileDefinition.query().filter_by(
        form=form,
        internal_name=internal_name,
    ).one()


@policy.context.post_authorize(action='view')
def value_definition_by_internal_name(form, internal_name) -> model.interface.InputValueDefinition:
    return model.interface.InputValueDefinition.query().filter_by(
        form=form,
        internal_name=internal_name,
    ).one()


def update_file_definition(definition: model.interface.InputFileDefinition, attributes: Dict):
    for attr in {'required', 'type', 'name', 'description', 'accept'} & set(attributes):
        setattr(definition, attr, attributes[attr])

    return definition


def update_value_definition(definition: model.interface.InputValueDefinition, attributes: Dict):
    for attr in {'required', 'type', 'name', 'description', 'default_binary'} & set(attributes):
        setattr(definition, attr, attributes[attr])

    if 'allowed_binaries' in attributes:
        current = set(definition.allowed_binaries)
        present = set(attributes['allowed_binaries'])
        removed = current - present
        if current != present:
            definition.allowed_binaries = [enum for enum in attributes['allowed_binaries'] if enum not in removed]
    elif definition.allowed_binaries:
        definition.allowed_binaries = []

    return definition


def job_executions_by_id(form_id, limit=50):
    form = form_by_id(form_id=form_id)
    return model.job.JobExecution.query().filter(
        (model.job.JobExecution.job_id == model.job.Job.id) &
        (model.job.Job.form_id == form.id)
    ).order_by(desc(model.job.JobExecution.updated)).limit(limit).all()



from celery import Celery
from riberry import config

app = Celery(main='background-tasks')

app.conf.update(config.config.celery)

app.conf.beat_schedule.update({
    'process:execution': {
        'task': 'riberry.celery.background.tasks.process_events',
        'schedule': config.config.background.events.interval,
        'kwargs': {
            'event_limit': config.config.background.events.processing_limit
        },
        'options': {'queue': 'riberry.background.events'}
    },
    'process:job-schedule': {
        'task': 'riberry.celery.background.tasks.job_schedules',
        'schedule': config.config.background.schedules.interval,
        'options': {'queue': 'riberry.background.schedules'}
    },
    'process:capacity': {
        'task': 'riberry.celery.background.tasks.update_capacity_parameters',
        'schedule': config.config.background.capacity.interval,
        'options': {'queue': 'riberry.background.schedules'}
    },
})

app.conf.imports = list(app.conf.imports) + ['riberry.celery.background.tasks']


def register_task(task_path, schedule):
    app.conf.beat_schedule[task_path] = {
        'task': 'riberry.celery.background.tasks.custom_task',
        'schedule': schedule,
        'args': [task_path],
        'options': {'queue': 'riberry.background.custom'}
    }



from typing import Union, Optional

import riberry
from riberry.model.job import ArtifactType


class Artifact:

    @staticmethod
    def create(
            filename: str,
            content: Union[bytes, str],
            name: str = None,
            type: Union[str, ArtifactType] = ArtifactType.output,
            category='Default',
            data: dict = None,
    ):
        return riberry.app.actions.artifacts.create_artifact(
            filename=filename,
            content=content,
            name=name,
            type=type,
            category=category,
            data=data,
        )

    @staticmethod
    def create_from_traceback(
            name: Optional[str] = None,
            filename: Optional[str] = None,
            category: str = 'Intercepted',
            type: riberry.model.job.ArtifactType = riberry.model.job.ArtifactType.error,
    ):
        return riberry.app.actions.artifacts.create_artifact_from_traceback(
            name=name,
            filename=filename,
            type=type,
            category=category,
        )



import uuid
from contextlib import contextmanager
from typing import Optional

import riberry


class ContextCurrent:

    def __init__(self, context):
        self.context: riberry.app.context.Context = context
        self._task_stream = None
        self._task_category = None
        self._task_step = None
        self._root_id = None
        self._task_id = None
        self._worker_uuid = str(uuid.uuid4())

    @property
    def WORKER_UUID(self):
        return self._worker_uuid

    @property
    def stream(self):
        return self._task_stream

    @property
    def step(self):
        return self._task_step

    @property
    def backend(self) -> 'riberry.app.backends.RiberryApplicationBackend':
        return self.riberry_app.backend

    @property
    def task(self):
        return self.backend.active_task()

    @property
    def riberry_app(self) -> 'riberry.app.RiberryApplication':
        return riberry.app.current_riberry_app

    @property
    def riberry_app_instance(self) -> riberry.model.application.ApplicationInstance:
        return riberry.app.env.get_instance_model()

    @property
    def task_id(self):
        return self._task_id

    @property
    def root_id(self):
        return self._root_id

    @property
    def job_execution(self) -> Optional[riberry.model.job.JobExecution]:
        return riberry.model.job.JobExecution.query().filter_by(task_id=self.root_id).first()

    @property
    def job(self) -> Optional[riberry.model.job.Job]:
        job_execution = self.job_execution
        return job_execution.job if self.job_execution else None

    @contextmanager
    def scope(self, root_id, task_id, stream, category, step):
        try:
            self._root_id, self._task_id = root_id, task_id
            self._task_stream, self._task_category, self._task_step = stream, category, step
            yield
        finally:
            self._root_id, self._task_id = None, None
            self._task_stream, self._task_category, self._task_step = None, None, None

    @property
    def progress(self) -> str:
        progress: riberry.model.job.JobExecutionProgress = riberry.model.job.JobExecutionProgress.query().filter_by(
            job_execution=self.job_execution,
        ).order_by(
            riberry.model.job.JobExecutionProgress.id.desc(),
        ).limit(
            1
        ).first()

        return progress.message if progress else None

    @progress.setter
    def progress(self, message: str):
        if message == self.progress:
            return

        progress = riberry.model.job.JobExecutionProgress(
            job_execution=self.job_execution,
            message=message,
        )

        riberry.model.conn.add(progress)
        riberry.model.conn.commit()



import os
from string import Template


def variable_substitution(obj):
    if isinstance(obj, str):
        try:
            return Template(template=obj).substitute(os.environ)
        except KeyError as exc:
            key = exc.args[0]
            raise ValueError(f'Environment variable substitution failed for {key!r}. '
                             f'Does the environment variable exist?')
    elif isinstance(obj, dict):
        return {k: variable_substitution(v) for k, v in obj.items()}
    elif isinstance(obj, (tuple, list, set)):
        return type(obj)(map(variable_substitution, obj))
    else:
        return obj



import types

import celery


def patch_send_task(instance: celery.Celery, func):
    send_task_original = instance.send_task
    def send_task(self, *args, **kwargs):
        func(self, *args, **kwargs)
        return send_task_original(*args, **kwargs)
    instance.send_task = types.MethodType(send_task, instance)




import math

import riberry
from .base import AddonStartStopStep
from celery.utils.log import logger as log


class Scale(riberry.app.addons.Addon):

    def __init__(
            self,
            active_parameter='active',
            concurrency_parameter='concurrency',
            minimum_concurrency=None,
            maximum_concurrency=None,
            check_queues=None,
    ):
        self.conf = ScaleConfiguration(
            active_parameter=active_parameter,
            concurrency_parameter=concurrency_parameter,
            minimum_concurrency=minimum_concurrency,
            maximum_concurrency=maximum_concurrency,
            check_queues=check_queues,
        )
        self.active_parameter = active_parameter
        self.concurrency_parameter = concurrency_parameter
        self.minimum_concurrency = minimum_concurrency
        self.maximum_concurrency = maximum_concurrency

    def register(self, riberry_app):
        class ConcreteScaleStep(ScaleStep):
            conf = self.conf
            rib = riberry_app

        riberry_app.backend.steps['worker'].add(ConcreteScaleStep)
        riberry_app.backend.user_options['worker'].add(self.regiser_user_options)

    @staticmethod
    def regiser_user_options(parser):
        parser.add_argument(
            '--rib-scale', action='store_true', default=False,
            help='Scale concurrency depending on activity (disabled by default)',
        )
        parser.add_argument(
            '--rib-scale-parameter',
            help='The name of the application instance parameter which stores the concurrency value',
        )
        parser.add_argument(
            '--rib-scale-group', default='default',
            help='Concurrency is distributed amongst all active workers in the same group',
        )
        parser.add_argument(
            '--rib-scale-max', default=None,
            help='Maximum concurrency when auto-scaling (uncapped by default)',
        )
        parser.add_argument(
            '--rib-scale-min', default=None,
            help='Minimum concurrency when auto-scaling (zero by default)',
        )

        feature_parser = parser.add_mutually_exclusive_group(required=False)
        feature_parser.add_argument(
            '--rib-scale-check-queues', dest='rib_scale_check_queues', action='store_true',
            help='Scale down if queues are empty and there are no active tasks'
        )
        feature_parser.add_argument(
            '--rib-scale-ignore-queues', dest='rib_scale_check_queues', action='store_false',
            help='Do not scale down if queues are empty and there are no active tasks'
        )
        parser.set_defaults(rib_scale_check_queues=True)


class ScaleConfiguration:

    def __init__(
            self,
            scale_concurrency=False,
            scale_group='default',
            active_parameter='active',
            concurrency_parameter='concurrency',
            minimum_concurrency=None,
            maximum_concurrency=None,
            check_queues=None,
    ):
        self.active_parameter = active_parameter
        self.concurrency_parameter = concurrency_parameter
        self.scale = scale_concurrency
        self.scale_group = scale_group
        self.minimum_concurrency = minimum_concurrency
        self.maximum_concurrency = maximum_concurrency
        self.check_queues = check_queues
        self.ignore_queues = set()


class ScaleStep(AddonStartStopStep):

    conf: ScaleConfiguration

    def __init__(self, worker, rib_scale, rib_scale_group, rib_scale_parameter, rib_scale_min, rib_scale_max, rib_scale_check_queues, **_):
        super().__init__(worker=worker, interval=1.0)

        self.conf.scale = bool(rib_scale)
        self.conf.scale_group = rib_scale_group or self.conf.scale_group
        self.conf.concurrency_parameter = rib_scale_parameter or self.conf.concurrency_parameter
        self.conf.minimum_concurrency = int(rib_scale_min) if rib_scale_min is not None else self.conf.minimum_concurrency
        self.conf.maximum_concurrency = int(rib_scale_max) if rib_scale_max is not None else self.conf.maximum_concurrency
        self.conf.check_queues = bool(rib_scale_check_queues) if rib_scale_check_queues is not None else self.conf.check_queues

        self.lock = riberry.app.util.redis_lock.RedisLock(name=f'step:scale:{self.conf.scale_group}', on_acquired=self.on_lock_acquired, interval=5000)

        self.queues = set()
        self.target_concurrency = None
        self.initial_concurrency = None
        self.is_active = False
        self._worker_uuid = self.rib.context.current.WORKER_UUID
        self._instance_name = self.rib.context.current.riberry_app_instance.internal_name
        self.idle_counter = 0

    def should_run(self) -> bool:
        return True

    def run(self):
        redis_instance = riberry.celery.util.celery_redis_instance()
        self.lock.run(redis_instance=redis_instance)
        if not self.consumer.task_consumer:
            return
        
        self.update(redis_instance=redis_instance)
        self.scale()

    def on_lock_acquired(self):
        r = riberry.celery.util.celery_redis_instance()
        epoch, _ = r.time()
        occurrence = set(r.zrevrangebyscore(name=self.scale_groups_log_key, max=epoch, min=epoch - 60))
        if occurrence:
            r.delete(self.scale_groups_active_temp_key)
            r.sadd(self.scale_groups_active_temp_key, *occurrence)
            r.rename(src=self.scale_groups_active_temp_key, dst=self.scale_groups_active_key)

    @classmethod
    def _tasks_available(cls, r, state, queues):
        return bool(
            len(state.reserved_requests) or
            len(state.active_requests) or
            state.requests or
            not cls._queues_empty(r, queues=queues)
        )

    @staticmethod
    def _queues_empty(r, queues):
        for queue_name in queues:
            for queue in r.keys(f'{queue_name}*'):
                try:
                    queue_length = r.llen(queue)
                    log.debug(f'ScaleStep:: Queue length of {queue.decode()!r} is {queue_length}')
                    if queue_length:
                        return False
                except:
                    continue
        return True

    def report(self, r):
        epoch, _ = r.time()
        r.zadd(name=self.scale_groups_log_key, mapping={self._worker_uuid: epoch})

    @property
    def scale_groups_active_key(self):
        return f'{self._instance_name}:scale-groups:{self.conf.scale_group}:active'

    @property
    def scale_groups_log_key(self):
        return f'{self._instance_name}:scale-groups:{self.conf.scale_group}:log'

    @property
    def scale_groups_active_temp_key(self):
        return f'{self._instance_name}:scale-groups:{self.conf.scale_group}:active-temp'

    def update(self, redis_instance):
        self.queues.update({q.name for q in self.worker.consumer.task_consumer.queues})
        if not self.initial_concurrency:
            self.initial_concurrency = self.worker.consumer.pool.num_processes

        instance = self.rib.context.current.riberry_app_instance
        active_flag = instance.active_schedule_value(name=self.conf.active_parameter, default='Y') == 'Y'
        tasks_available = self._tasks_available(r=redis_instance, state=self.worker_state, queues=self.queues - self.conf.ignore_queues)
        if tasks_available:
            self.report(r=redis_instance)

        self.is_active = active_flag and (tasks_available if self.conf.check_queues else True)
        if not self.is_active:
            if self.idle_counter > 10 or self.target_concurrency is None:
                self.target_concurrency = 0
            self.idle_counter += 1
            return
        self.idle_counter = 0

        scale_group = list(sorted(b.decode() for b in redis_instance.smembers(self.scale_groups_active_key)))
        if self.conf.scale and self.conf.concurrency_parameter:
            target_concurrency = instance.active_schedule_value(name=self.conf.concurrency_parameter, default=None)
            if target_concurrency is None:
                target_concurrency = self.initial_concurrency
            else:
                target_concurrency = int(target_concurrency)
            target_concurrency *= (1 / len(scale_group)) if self._worker_uuid in scale_group else 0

            if target_concurrency:
                if scale_group.index(self._worker_uuid) == 0:
                    target_concurrency = math.ceil(target_concurrency)
                else:
                    target_concurrency = math.floor(target_concurrency)
        else:
            target_concurrency = self.initial_concurrency

        if self.conf.maximum_concurrency is not None:
            target_concurrency = min(target_concurrency, self.conf.maximum_concurrency)

        if self.conf.minimum_concurrency is not None:
            target_concurrency = max(target_concurrency, self.conf.minimum_concurrency)

        target_concurrency = int(target_concurrency)
        if target_concurrency != self.target_concurrency:
            self.target_concurrency = target_concurrency

    def scale(self):
        actual_concurrency = self.worker.consumer.pool.num_processes
        target_concurrency = self.target_concurrency

        scale_group = list(sorted(b.decode() for b in riberry.celery.util.celery_redis_instance().smembers(self.scale_groups_active_key)))
        log.debug(f'A: {self.is_active} C[T]: {self.target_concurrency}, C[A]: {actual_concurrency}, P: {self.worker.consumer.qos.value},  M: {scale_group}')

        if target_concurrency == 0:
            for queue in list(self.consumer.task_consumer.queues):
                if queue.name not in self.conf.ignore_queues:
                    self.consumer.cancel_task_queue(queue)
        else:
            queue_names = {q.name for q in self.consumer.task_consumer.queues}
            for queue in self.queues:
                if queue not in queue_names and queue not in self.conf.ignore_queues:
                    self.consumer.add_task_queue(queue)

        if target_concurrency > actual_concurrency:
            if actual_concurrency == 0:
                self.worker.consumer.pool.grow(1)
            else:
                self.worker.consumer.pool.grow(min(target_concurrency - actual_concurrency, 8))
            log.info(f'ScaleStep:: Scaled concurrency up to {self.worker.consumer.pool.num_processes} concurrency (target: {self.target_concurrency}, prefetch: {self.worker.consumer.qos.value})')
        elif actual_concurrency > target_concurrency:
            self.worker.consumer.qos.decrement_eventually(n=1)
            self.worker.consumer.qos.update()
            self.worker.consumer.pool.shrink(min(actual_concurrency - target_concurrency, 8))
            log.info(f'ScaleStep:: Scaled concurrency down to {self.worker.consumer.pool.num_processes} concurrency (target: {self.target_concurrency}, prefetch: {self.worker.consumer.qos.value})')

        prefetch_target = self.worker.consumer.pool.num_processes * self.worker.consumer.prefetch_multiplier
        prefetch_difference = prefetch_target - self.worker.consumer.qos.value
        if prefetch_difference > 0:
            self.worker.consumer.qos.increment_eventually(n=abs(prefetch_difference))
            self.worker.consumer.qos.update()
            log.info(f'ScaleStep:: Scaled prefetch up to {self.worker.consumer.qos.value} (+{prefetch_difference})')
        elif prefetch_difference < 0 and self.worker.consumer.qos.value != 1:
            self.worker.consumer.qos.decrement_eventually(n=abs(prefetch_difference))
            self.worker.consumer.qos.update()
            log.info(f'ScaleStep:: Scaled prefetch down to {self.worker.consumer.qos.value} ({prefetch_difference})')




import json
from typing import Optional
import uuid

import riberry


def create_external_task(
        job_execution: riberry.model.job.JobExecution,
        name: str = None,
        type: str = 'external',
        external_task_id: Optional[str] = None,
        input_data: Optional[bytes] = None
):
    if external_task_id is None:
        external_task_id = str(uuid.uuid4())

    if name is None:
        name = external_task_id

    if isinstance(input_data, str):
        input_data = input_data.encode()

    if input_data is not None and not isinstance(input_data, bytes):
        input_data = json.dumps(input_data).encode()

    external_task = riberry.model.job.JobExecutionExternalTask(
        job_execution=job_execution,
        stream_id=None,
        task_id=external_task_id,
        name=name,
        type=type,
        input_data=input_data,
    )

    riberry.model.conn.add(external_task)
    riberry.model.conn.commit()

    return external_task


def mark_as_ready(external_task_id, output_data):
    task: riberry.model.job.JobExecutionExternalTask = riberry.model.job.JobExecutionExternalTask.query().filter_by(
        task_id=external_task_id,
    ).one()

    if isinstance(output_data, str):
        output_data = output_data.encode()

    if output_data is not None and not isinstance(output_data, bytes):
        output_data = json.dumps(output_data).encode()

    task.status = 'READY'
    task.output_data = output_data
    riberry.model.conn.commit()



import time

from redis.exceptions import LockError


class RedisLock:

    def __init__(self, name, on_acquired, interval, min_interval=100):
        self.name = name
        self.on_acquired = on_acquired
        self.interval = interval
        self.min_interval = min_interval

    @property
    def key_timeout(self):
        return f'lock:{self.name}:timeout'

    @property
    def key_lock(self):
        return f'lock:{self.name}:acquire'

    def run(self, redis_instance):

        if redis_instance.get(name=self.key_timeout) is None:
            try:
                self._attempt_lock(redis_instance=redis_instance)
            except LockError:
                pass

    def _attempt_lock(self, redis_instance):
        with redis_instance.lock(name=self.key_lock, timeout=60, blocking_timeout=0.0):
            print(f"{self.name}: acquired lock...")
            time_start = time.time()
            try:
                self.on_acquired()
            finally:
                process_time = time.time() - time_start
                self._set_timeout(redis_instance=redis_instance, process_time=process_time)

    def _set_timeout(self, redis_instance, process_time):
        expiry = int(max(self.interval - (process_time * 1000), self.min_interval))
        redis_instance.set(name=self.key_timeout, value='1', px=expiry)
        print(f'{self.name}: processed task in {process_time:03}, setting lock expiry to {expiry:03} milliseconds')
